# ----------------------------------------------------------------------
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------
import re

from apparmor.common import AppArmorException, convert_regexp, open_file_read, warn  # , debug, error, msg


class Severity:
    def __init__(self, dbname=None, default_rank=10):
        """Initialises the class object"""
        self.PROF_DIR = '/etc/apparmor.d'  # The profile directory
        self.NOT_IMPLEMENTED = '_-*not*implemented*-_'  # used for rule types that don't have severity ratings
        self.severity = dict()
        self.severity['DATABASENAME'] = dbname
        self.severity['CAPABILITIES'] = {}
        self.severity['FILES'] = {}
        self.severity['REGEXPS'] = {}
        self.severity['DEFAULT_RANK'] = default_rank
        # For variable expansions for the profile
        self.severity['VARIABLES'] = dict()
        if not dbname:
            raise AppArmorException("No severity db file given")

        with open_file_read(dbname) as database:  # open(dbname, 'r')
            for lineno, line in enumerate(database, start=1):
                line = line.strip()  # or only rstrip and lstrip?
                if not line or line.startswith('#'):
                    continue
                if line.startswith('/'):
                    try:
                        path, read, write, execute = line.split()
                        read, write, execute = int(read), int(write), int(execute)
                    except ValueError:
                        raise AppArmorException(
                            "Insufficient values for permissions in file: %s\n\t[Line %s]: %s"
                            % (dbname, lineno, line))
                    else:
                        if read not in range(0, 11) or write not in range(0, 11) or execute not in range(0, 11):
                            raise AppArmorException(
                                "Inappropriate values for permissions in file: %s\n\t[Line %s]: %s"
                                % (dbname, lineno, line))
                        path = path.lstrip('/')
                        if '*' not in path:
                            self.severity['FILES'][path] = {'r': read, 'w': write, 'x': execute}
                        else:
                            ptr = self.severity['REGEXPS']
                            pieces = path.split('/')
                            for index, piece in enumerate(pieces):  # pragma: no branch
                                if '*' in piece:
                                    path = '/'.join(pieces[index:])
                                    regexp = convert_regexp(path)
                                    ptr[regexp] = {'AA_RANK': {'r': read, 'w': write, 'x': execute}}
                                    break
                                else:
                                    ptr[piece] = ptr.get(piece, {})
                                    ptr = ptr[piece]
                elif line.startswith('CAP_'):
                    try:
                        resource, severity = line.split()
                        severity = int(severity)
                    except ValueError:
                        error_message = 'No severity value present in file: %s\n\t[Line %s]: %s' % (dbname, lineno, line)
                        # error(error_message)
                        raise AppArmorException(error_message)  # from None
                    else:
                        if severity not in range(0, 11):
                            raise AppArmorException(
                                "Inappropriate severity value present in file: %s\n\t[Line %s]: %s"
                                % (dbname, lineno, line))
                        self.severity['CAPABILITIES'][resource] = severity
                else:
                    raise AppArmorException("Unexpected line in file: %s\n\t[Line %s]: %s" % (dbname, lineno, line))

    def rank_capability(self, resource):
        """Returns the severity of for the capability resource, default value if no match"""
        cap = 'CAP_%s' % resource.upper()
        if resource == '__ALL__':
            return max(self.severity['CAPABILITIES'].values())
        if cap in self.severity['CAPABILITIES'].keys():
            return self.severity['CAPABILITIES'][cap]
        # raise ValueError("unexpected capability rank input: %s"%resource)
        warn("unknown capability: %s" % resource)
        return self.severity['DEFAULT_RANK']

    def rank_path(self, path, mode=None):
        """Returns the rank for the given path"""
        if '@' in path:    # path contains variable
            return self.handle_variable_rank(path, mode)
        elif path.startswith('/'):    # file resource
            return self.handle_file(path, mode)
        else:
            raise AppArmorException("Unexpected path input: %s" % path)

    def check_subtree(self, tree, mode, sev, segments):
        """Returns the max severity from the regex tree"""
        if not segments:
            first = ''
        else:
            first = segments[0]
        rest = segments[1:]
        path = '/'.join([first] + rest)
        # Check if we have a matching directory tree to descend into
        if tree.get(first, False):
            sev = self.check_subtree(tree[first], mode, sev, rest)
        # If severity still not found, match against globs
        if sev is None:
            # Match against all globs at this directory level
            for chunk in tree.keys():
                if '*' in chunk:
                    # Match rest of the path
                    if re.search("^" + chunk, path):
                        # Find max rank
                        if "AA_RANK" in tree[chunk].keys():  # pragma: no branch
                            for m in mode:
                                if sev is None or tree[chunk]["AA_RANK"].get(m, -1) > sev:
                                    sev = tree[chunk]["AA_RANK"].get(m, None)
        return sev

    def handle_file(self, resource, mode):
        """Returns the severity for the file, default value if no match found"""
        resource = resource[1:]    # remove initial / from path
        pieces = resource.split('/')    # break path into directory level chunks
        sev = None
        # Check for an exact match in the db
        if resource in self.severity['FILES'].keys():
            # Find max value among the given modes
            for m in mode:
                if sev is None or self.severity['FILES'][resource].get(m, -1) > sev:
                    sev = self.severity['FILES'][resource].get(m, None)
        else:
            # Search regex tree for matching glob
            sev = self.check_subtree(self.severity['REGEXPS'], mode, sev, pieces)
        if sev is None:
            # Return default rank if severity cannot be found
            return self.severity['DEFAULT_RANK']
        else:
            return sev

    def handle_variable_rank(self, resource, mode):
        """Returns the max possible rank for file resources containing variables"""
        regex_variable = re.compile('@{([^{.]*)}')
        matches = regex_variable.search(resource)
        if matches:
            rank = self.severity['DEFAULT_RANK']
            variable = '@{%s}' % matches.groups()[0]
            # variables = regex_variable.findall(resource)
            for replacement in self.severity['VARIABLES'][variable]:
                resource_replaced = self.variable_replace(variable, replacement, resource)
                rank_new = self.handle_variable_rank(resource_replaced, mode)
                if rank == self.severity['DEFAULT_RANK']:
                    rank = rank_new
                elif rank_new != self.severity['DEFAULT_RANK'] and rank_new > rank:
                    rank = rank_new
            return rank
        else:
            return self.handle_file(resource, mode)

    def variable_replace(self, variable, replacement, resource):
        """Returns the expanded path for the passed variable"""
        leading = False
        trailing = False
        # Check for leading or trailing / that may need to be collapsed
        if resource.find("/" + variable) != -1 and resource.find("//" + variable) == -1:  # find that a single / exists before variable or not
            leading = True
        if resource.find(variable + "/") != -1 and resource.find(variable + "//") == -1:
            trailing = True
        if replacement.startswith('/') and not replacement.startswith('//') and leading:  # finds if the replacement has leading / or not
            replacement = replacement[1:]
        if replacement.endswith('/') and not replacement.endswith('//') and trailing:
            replacement = replacement[:-1]
        return resource.replace(variable, replacement)

    def set_variables(self, vars):
        """Set the profile variables to use for rating the severity"""
        self.severity['VARIABLES'] = vars



# ------------------------------------------------------------------
#
#    Copyright (C) 2012 Canonical Ltd.
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License published by the Free Software Foundation.
#
# ------------------------------------------------------------------

import collections
import logging
import os
import re
import subprocess
import signal
import sys
import termios
import tty
from tempfile import NamedTemporaryFile

DEBUGGING = False


#
# Utility classes
#
class AppArmorException(Exception):
    """This class represents AppArmor exceptions"""
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class AppArmorBug(Exception):
    """This class represents AppArmor exceptions "that should never happen"."""


#
# Utility functions
#
def error(out, exit_code=1, do_exit=True):
    """Print error message and exit"""
    try:
        print("ERROR: %s" % (out), file=sys.stderr)
    except IOError:
        pass

    if do_exit:
        sys.exit(exit_code)


def warn(out):
    """Print warning message"""
    try:
        print("WARN: %s" % (out), file=sys.stderr)
    except IOError:
        pass


def msg(out, output=sys.stdout):
    """Print message"""
    try:
        print("%s" % (out), file=output)
    except IOError:
        pass


def debug(out):
    """Print debug message"""
    if DEBUGGING:
        try:
            print("DEBUG: %s" % (out), file=sys.stderr)
        except IOError:
            pass


def recursive_print(src, dpth=0, key=''):
    # print recursively in a nicely formatted way
    # useful for debugging, too verbose for production code ;-)

    # based on code "stolen" from Scott S-Allen / MIT License
    # http://code.activestate.com/recipes/578094-recursively-print-nested-dictionaries/
    """Recursively prints nested elements."""
    tabs = ' ' * dpth * 4  # or 2 or 8 or...

    if isinstance(src, dict):
        empty = True
        for key in src.keys():
            print(tabs + '[%s]' % key)
            recursive_print(src[key], dpth + 1, key)
            empty = False
        if empty:
            print(tabs + '[--- empty ---]')
    elif isinstance(src, list) or isinstance(src, tuple):
        if not src:
            print(tabs + '[--- empty ---]')
        else:
            print(tabs + "[")
            for litem in src:
                recursive_print(litem, dpth + 1)
            print(tabs + "]")
    else:
        if key:
            print(tabs + '%s = %s' % (key, src))
        else:
            print(tabs + '- %s' % src)


def subprocess_setup():
    # Python installs a SIGPIPE handler by default. This is usually not what
    # non-Python subprocesses expect.
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)


def cmd_pipe_stderr(command, stdin=None):
    return cmd(command, stderr=subprocess.PIPE, output_fmt='utf-8', stdin=stdin)


def cmd(command, stderr=subprocess.STDOUT, output_fmt='ascii', stdin=None):
    """Execute the given command and return its output as a tuple (returncode, output)."""
    debug(command)
    try:
        sp = subprocess.Popen(command, stdin=stdin, stdout=subprocess.PIPE, stderr=stderr, close_fds=True, preexec_fn=subprocess_setup)
    except OSError as ex:
        return [127, str(ex)]

    out, err = sp.communicate()  # TODO input???

    if err:
        out = err

    if output_fmt == 'ascii':
        decoded = out.decode(output_fmt, 'ignore')
    else:
        decoded = out.decode(output_fmt)

    return [sp.returncode, decoded]


def cmd_pipe(command1, command2):
    """Try to pipe command1 into command2."""
    try:
        sp1 = subprocess.Popen(command1, stdout=subprocess.PIPE)
        sp2 = subprocess.Popen(command2, stdin=sp1.stdout)
    except OSError as ex:
        return [127, str(ex)]

    out = sp2.communicate()[0].decode('ascii', 'ignore')

    return [sp2.returncode, out]


def valid_path(path):
    """Valid path"""
    m = "Invalid path: %s" % (path)
    if not path.startswith('/'):  # No relative paths
        debug("%s (relative)" % (m))
        return False

    if '"' in path:  # We double quote elsewhere
        debug("%s (contains quote)" % (m))
        return False

    return True


def is_skippable_file(path):
    """Returns True if filename matches something to be skipped (rpm or dpkg backup files, hidden files etc.)
        The list of skippable files needs to be synced with apparmor initscript and libapparmor _aa_is_blacklisted()
        path: filename (with or without directory)"""

    basename = os.path.basename(path)

    if not basename or basename.startswith('.') or basename == 'README':
        return True

    skippable_suffix = (
        '.dpkg-new', '.dpkg-old', '.dpkg-dist', '.dpkg-bak', '.dpkg-remove',
        '.pacsave', '.pacnew', '.rpmnew', '.rpmsave', '.orig', '.rej', '~')
    if basename.endswith(skippable_suffix):
        return True

    return False


def open_file_read(path, encoding='UTF-8'):
    """Open specified file read-only"""
    return open_file_anymode('r', path, encoding)


def open_file_write(path):
    """Open specified file in write/overwrite mode"""
    return open_file_anymode('w', path, 'UTF-8')


def open_file_anymode(mode, path, encoding='UTF-8'):
    """Crash-resistant wrapper to open a specified file in specified mode"""

    # This avoids a crash when reading a logfile with special characters that
    # are not utf8-encoded (for example a latin1 "ö"), and also avoids crashes
    # at several other places we don't know yet ;-)
    return open(path, mode, encoding=encoding, errors='surrogateescape')


def readkey():
    """Returns the pressed key"""
    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(sys.stdin.fileno())
        ch = sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

    return ch


def hasher():
    """A neat alternative to perl's hash reference"""
    # Creates a dictionary for any depth and returns empty dictionary otherwise
    # WARNING: when reading non-existing sub-dicts, empty dicts will be added.
    #          This might cause strange effects when using .keys()
    return collections.defaultdict(hasher)


def convert_regexp(regexp):
    regex_paren = re.compile('^(.*){([^}]*)}(.*)$')
    regexp = regexp.strip()

    regexp = regexp.replace('(', '\\(').replace(')', '\\)')  # escape '(' and ')'

    new_reg = re.sub(r'(?<!\\)([.+$])', r'\\\1', regexp)

    while regex_paren.search(new_reg):
        match = regex_paren.search(new_reg).groups()
        prev = match[0]
        after = match[2]
        p1 = match[1].replace(',', '|')
        new_reg = prev + '(' + p1 + ')' + after

    new_reg = new_reg.replace('?', '[^/\000]')

    multi_glob = '__KJHDKVZH_AAPROF_INTERNAL_GLOB_SVCUZDGZID__'
    new_reg = new_reg.replace('**', multi_glob)
    # print(new_reg)

    # Match at least one character if * or ** after /
    # ?< is the negative lookback operator
    new_reg = new_reg.replace('*', '(((?<=/)[^/\000]+)|((?<!/)[^/\000]*))')
    new_reg = new_reg.replace(multi_glob, '(((?<=/)[^\000]+)|((?<!/)[^\000]*))')
    if not regexp.startswith('^'):
        new_reg = '^' + new_reg
    if not regexp.endswith('$'):
        new_reg = new_reg + '$'
    return new_reg


def user_perm(prof_dir):
    if not os.access(prof_dir, os.W_OK):
        sys.stdout.write("Cannot write to profile directory.\n"
                         "Please run as a user with appropriate permissions.\n")
        return False
    return True


def split_name(full_profile):
    if '//' in full_profile:
        profile, hat = full_profile.split('//')[:2]  # XXX limit to two levels to avoid an Exception on nested child profiles or nested null-*
        # TODO: support nested child profiles
    else:
        profile = full_profile
        hat = full_profile

    return (profile, hat)


def combine_profname(name_parts):
    """combine name_parts (main profile, child) into a joint main//child profile name"""

    if not isinstance(name_parts, (list, tuple)):
        raise AppArmorBug('combine_profname() called with parameter of type %s, must be a list or tuple' % type(name_parts))

    # if last item is None, drop it (can happen when called with [profile, hat] when hat is None)
    if name_parts[-1] is None:
        name_parts = name_parts[:-1]

    return '//'.join(name_parts)


class DebugLogger:
    """Unified debug facility. Logs to file or stderr.

    Does not log anything by default. Will only log if environment variable
    LOGPROF_DEBUG is set to a number between 1 and 3 or if method activateStderr
    is run.
    """
    def __init__(self, module_name=__name__):
        self.debugging = False
        self.debug_level = logging.DEBUG

        if os.getenv('LOGPROF_DEBUG', False):
            self.logfile = '/var/log/apparmor/logprof.log'
            self.debugging = os.getenv('LOGPROF_DEBUG')
            try:
                self.debugging = int(self.debugging)
            except (TypeError, ValueError):
                self.debugging = False
            if self.debugging not in range(0, 4):
                sys.stdout.write('Environment Variable: LOGPROF_DEBUG contains invalid value: %s'
                                 % os.getenv('LOGPROF_DEBUG'))
            if self.debugging == 0:  # debugging disabled, don't need to setup logging
                return
            if self.debugging == 1:
                self.debug_level = logging.ERROR  # 40
            elif self.debugging == 2:
                self.debug_level = logging.INFO  # 20
            elif self.debugging == 3:
                self.debug_level = logging.DEBUG  # 10

            try:
                logging.basicConfig(filename=self.logfile, level=self.debug_level,
                                    format='%(asctime)s - %(name)s - %(message)s\n')
            except IOError:
                # Unable to open the default logfile, so create a temporary logfile and tell use about it
                templog = NamedTemporaryFile('w', prefix='apparmor', suffix='.log', delete=False)
                templog.close()
                sys.stdout.write("\nCould not open: %s\nLogging to: %s\n" % (self.logfile, templog.name))

                logging.basicConfig(filename=templog.name, level=self.debug_level,
                                    format='%(asctime)s - %(name)s - %(message)s\n')

            self.logger = logging.getLogger(module_name)

    def activateStderr(self):
        self.debugging = True
        logging.basicConfig(
            level=self.debug_level,
            format='%(levelname)s: %(message)s',
            stream=sys.stderr,
        )
        self.logger = logging.getLogger(__name__)

    def error(self, message, *args):
        if self.debugging:
            self.logger.error(message, *args)

    def info(self, message, *args):
        if self.debugging:
            self.logger.info(message, *args)

    def debug(self, message, *args):
        if self.debugging:
            self.logger.debug(message, *args)

    def shutdown(self):
        logging.shutdown()
        # logging.shutdown([self.logger])



# ----------------------------------------------------------------------
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#    Copyright (C) 2015 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import re

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.regex import RE_PROFILE_NETWORK, strip_parenthesis, aare
from apparmor.rule import BaseRule, BaseRuleset, logprof_value_or_all, parse_modifiers, check_dict_keys, \
    check_and_split_list, initialize_cond_dict, print_dict_values, tuple_to_dict
from apparmor.translations import init_translation
from apparmor.rule.unix import unix_accesses as network_accesses
from apparmor.rule.unix import access_flags
import ipaddress

_ = init_translation()

network_domain_keywords = [
    'unspec', 'unix', 'inet', 'ax25', 'ipx', 'appletalk', 'netrom', 'bridge', 'atmpvc', 'x25', 'inet6',
    'rose', 'netbeui', 'security', 'key', 'netlink', 'packet', 'ash', 'econet', 'atmsvc', 'rds', 'sna',
    'irda', 'pppox', 'wanpipe', 'llc', 'ib', 'mpls', 'can', 'tipc', 'bluetooth', 'iucv', 'rxrpc', 'isdn',
    'phonet', 'ieee802154', 'caif', 'alg', 'nfc', 'vsock', 'kcm', 'qipcrtr', 'smc', 'xdp', 'mctp']

network_type_keywords = ['stream', 'dgram', 'seqpacket', 'rdm', 'raw', 'packet']
network_protocol_keywords = ['tcp', 'udp', 'icmp']

byte = r'(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)'
network_ipv4 = fr'{byte}\.{byte}\.{byte}\.{byte}'

network_ipv6 = (
    r'('
    r'([0-9a-f]{1,4}:){7}[0-9a-f]{1,4}|'
    r'([0-9a-f]{1,4}:){1,7}:|'
    r'([0-9a-f]{1,4}:){1,6}:[0-9a-f]{1,4}|'
    r'([0-9a-f]{1,4}:){1,5}(:[0-9a-f]{1,4}){1,2}|'
    r'([0-9a-f]{1,4}:){1,4}(:[0-9a-f]{1,4}){1,3}|'
    r'([0-9a-f]{1,4}:){1,3}(:[0-9a-f]{1,4}){1,4}|'
    r'([0-9a-f]{1,4}:){1,2}(:[0-9a-f]{1,4}){1,5}|'
    r'[0-9a-f]{1,4}:((:[0-9a-f]{1,4}){1,6})|'
    r':((:[0-9a-f]{1,4}){1,7}|:)|'
    r'fe80:(:[0-9a-f]{0,4}){0,4}%%[0-9a-zA-Z]{1,}|'
    r'::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|'
    r'([0-9a-f]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])'
    r')(%%[0-9a-zA-Z]{1,})?'
)

network_port = r'(port\s*=\s*(?P<%s>\d+))[,\s]*'
ip_cond = fr'\s*ip\s*=\s*(?P<%s>(({network_ipv4})|({network_ipv6})|none))[,\s]*'
network_interface = r'(interface\s*=\s*(?P<%s>' + aare + r'))[,\s]*'
network_label = r'(label\s*=\s*(?P<%s>' + aare + r'))[,\s]*'

RE_LOCAL_EXPR = f'((({ip_cond % "ip"})|({network_port % "port"})|({network_interface % "interface"})|({network_label % "label"}))*)'
RE_PEER_EXPR = fr'(peer\s*=\s*\(\s*(({ip_cond % "ip_peer"})|({network_port % "port_peer"})|({network_interface % "interface_peer"})|({network_label % "label_peer"}))+\s*\))'


RE_NETWORK_DOMAIN = '(' + '|'.join(network_domain_keywords) + ')'
RE_NETWORK_TYPE = '(' + '|'.join(network_type_keywords) + ')'
RE_NETWORK_PROTOCOL = '(' + '|'.join(network_protocol_keywords) + ')'

RE_NETWORK_DETAILS = re.compile(
    r'^\s*'
    + r'(\s*' + network_accesses + r')?\s*'
    + '(?P<domain>' + RE_NETWORK_DOMAIN + r')?\s*'  # optional domain
    + r'(\s+(?P<type_or_protocol>' + RE_NETWORK_TYPE + '|' + RE_NETWORK_PROTOCOL + r'))?\s*'  # optional type or protocol
    + '(' + RE_LOCAL_EXPR + r')?\s*'
    + '(' + RE_PEER_EXPR + r')?\s*'
    + r'$')

non_peer_accesses = {'create', 'bind', 'listen', 'shutdown', 'getattr', 'setattr', 'getopt', 'setopt'}


class NetworkRule(BaseRule):
    """Class to handle and store a single network rule"""

    # Nothing external should reference this class, all external users
    # should reference the class field NetworkRule.ALL
    class __NetworkAll:
        pass

    ALL = __NetworkAll

    rule_name = 'network'
    _match_re = RE_PROFILE_NETWORK

    def __init__(self, accesses, domain, type_or_protocol, local_expr, peer_expr, audit=False, deny=False,
                 allow_keyword=False, comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny, allow_keyword=allow_keyword,
                         comment=comment, log_event=log_event, priority=priority)

        if type(local_expr) is tuple:
            if accesses is None:
                accesses = self.ALL
            else:
                accesses = accesses.split()
            local_expr = tuple_to_dict(local_expr, ['ip', 'port', 'interface', 'label'])
            peer_expr = tuple_to_dict(peer_expr, ['ip', 'port', 'interface', 'label'])

        self.accesses, self.all_accesses, unknown_items = check_and_split_list(accesses, access_flags, self.ALL,  type(self).__name__, 'accesses')

        if unknown_items:
            raise AppArmorException(f'Invalid access in Network rule: {unknown_items}')

        self.local_expr = check_dict_keys(local_expr, {'ip', 'port', 'interface', 'label'}, self.ALL)
        self.peer_expr = check_dict_keys(peer_expr, {'ip', 'port', 'interface', 'label'}, self.ALL)

        if self.local_expr != self.ALL and 'port' in self.local_expr and int(self.local_expr['port']) > 65535:
            raise AppArmorException(f'Invalid port: {self.local_expr["port"]}')
        if self.peer_expr != self.ALL and 'port' in self.peer_expr and int(self.peer_expr['port']) > 65535:
            raise AppArmorException(f'Invalid remote port: {self.peer_expr["port"]}')

        if self.local_expr != self.ALL and 'ip' in self.local_expr and not is_valid_ip(self.local_expr['ip']):
            raise AppArmorException(f'Invalid ip: {self.local_expr["ip"]}')
        if self.peer_expr != self.ALL and 'ip' in self.peer_expr and not is_valid_ip(self.peer_expr['ip']):
            raise AppArmorException(f'Invalid ip: {self.peer_expr["ip"]}')

        if not self.all_accesses and self.peer_expr != self.ALL and self.accesses & non_peer_accesses:
            raise AppArmorException('Cannot use a peer_expr and an access in the set (%s) simultaneously' % ', '.join(non_peer_accesses))

        self.domain = None
        self.all_domains = False
        if domain == self.ALL:
            self.all_domains = True
        elif isinstance(domain, str):
            if domain in network_domain_keywords:
                self.domain = domain

                if not self.domain.startswith('inet') and (self.local_expr != self.ALL or self.peer_expr != self.ALL):
                    raise AppArmorException('Cannot use a local expression or a peer expression for non-inet domains')
            else:
                raise AppArmorBug('Passed unknown domain to %s: %s' % (type(self).__name__, domain))
        else:
            raise AppArmorBug('Passed unknown object to %s: %s' % (type(self).__name__, str(domain)))

        self.type_or_protocol = None
        self.all_type_or_protocols = False
        if type_or_protocol == self.ALL:
            self.all_type_or_protocols = True
        elif isinstance(type_or_protocol, str):
            if type_or_protocol in network_protocol_keywords:
                self.type_or_protocol = type_or_protocol
            elif type_or_protocol in network_type_keywords:
                self.type_or_protocol = type_or_protocol
            else:
                raise AppArmorBug('Passed unknown type_or_protocol to %s: %s' % (type(self).__name__, type_or_protocol))
        else:
            raise AppArmorBug('Passed unknown object to %s: %s' % (type(self).__name__, str(type_or_protocol)))

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        """parse raw_rule and return instance of this class"""

        priority, audit, deny, allow_keyword, comment = parse_modifiers(matches)

        rule_details = ''
        if matches.group('details'):
            rule_details = matches.group('details')

        if rule_details:
            details = RE_NETWORK_DETAILS.search(rule_details)
            if not details:
                raise AppArmorException(_("Invalid or unknown keywords in 'network %s'" % rule_details))

            r = details.groupdict()

            domain = r['domain'] or cls.ALL
            type_or_protocol = r['type_or_protocol'] or cls.ALL

            if r['accesses']:
                accesses = strip_parenthesis(r['accesses']).replace(',', ' ').split()
            else:
                accesses = cls.ALL

            local_expr = initialize_cond_dict(r, ['ip', 'port', 'interface', 'label'], '', cls.ALL)
            peer_expr = initialize_cond_dict(r, ['ip', 'port', 'interface', 'label'], '_peer', cls.ALL)

        else:
            accesses = cls.ALL
            domain = cls.ALL
            type_or_protocol = cls.ALL
            local_expr = cls.ALL
            peer_expr = cls.ALL

        return cls(accesses, domain, type_or_protocol, local_expr, peer_expr,
                   audit=audit, deny=deny, allow_keyword=allow_keyword, comment=comment, priority=priority)

    def get_clean(self, depth=0):
        """return rule (in clean/default formatting)"""

        space = '  ' * depth

        accesses = ' (%s)' % (', '.join(sorted(self.accesses))) if not self.all_accesses else ''

        if self.all_domains:
            domain = ''
        elif self.domain:
            domain = ' %s' % self.domain
        else:
            raise AppArmorBug('Empty domain in network rule')

        if self.all_type_or_protocols:
            type_or_protocol = ''
        elif self.type_or_protocol:
            type_or_protocol = ' %s' % self.type_or_protocol
        else:
            raise AppArmorBug('Empty type or protocol in network rule')

        local_expr = print_dict_values(self.local_expr, self.ALL)
        peer_expr = print_dict_values(self.peer_expr, self.ALL, 'peer')

        return ('%s%snetwork%s%s%s%s%s,%s' % (space, self.modifiers_str(), accesses, domain, type_or_protocol, local_expr, peer_expr, self.comment))

    def _is_covered_localvars(self, other_rule):
        """check if other_rule is covered by this rule object"""

        if not self._is_covered_list(self.accesses, self.all_accesses, other_rule.accesses, other_rule.all_accesses, 'accesses'):
            return False

        if not self._is_covered_plain(self.domain, self.all_domains, other_rule.domain, other_rule.all_domains, 'domain'):
            return False

        if not self._is_covered_plain(self.type_or_protocol, self.all_type_or_protocols, other_rule.type_or_protocol, other_rule.all_type_or_protocols, 'type or protocol'):
            return False

        if not self._is_covered_dict(self.local_expr, other_rule.local_expr):
            return False
        if not self._is_covered_dict(self.peer_expr, other_rule.peer_expr):
            return False

        # still here? -> then it is covered
        return True

    def _is_covered_dict(self, d, other):

        if d is self.ALL:
            return True
        elif other is self.ALL:
            return False

        for it in other:
            if it not in d:
                continue  # No constraints on this item.
            else:
                if not self._is_covered_plain(d[it], False, other[it], False, it):
                    return False

        return True

    def _is_equal_localvars(self, other_rule, strict):
        """compare if rule-specific variables are equal"""

        if self.accesses != other_rule.accesses:
            return False

        if (self.domain != other_rule.domain
                or self.all_domains != other_rule.all_domains):
            return False

        if (self.type_or_protocol != other_rule.type_or_protocol
                or self.all_type_or_protocols != other_rule.all_type_or_protocols):
            return False

        if self.local_expr != other_rule.local_expr:
            return False

        if self.peer_expr != other_rule.peer_expr:
            return False

        return True

    def _logprof_header_localvars(self):
        accesses = logprof_value_or_all(self.accesses, self.all_accesses)
        family    = logprof_value_or_all(self.domain,           self.all_domains)  # noqa: E221
        sock_type = logprof_value_or_all(self.type_or_protocol, self.all_type_or_protocols)

        local_expr = logprof_value_or_all(self.local_expr, self.local_expr == self.ALL)
        peer_expr = logprof_value_or_all(self.peer_expr, self.peer_expr == self.ALL)

        return (
            _('Accesses'), accesses,
            _('Network Family'), family,
            _('Socket Type'), sock_type,
            _('Local'), local_expr,
            _('Peer'), peer_expr,
        )

    # XXX: interface and peer_interface are not present in kernel log events yet.
    @staticmethod
    def hashlog_from_event(hl, ev):
        local = (ev.get('addr'), ev.get('port'), ev.get('interface'), ev.get('label'))
        peer = (ev.get('peer_addr'), ev.get('remote_port'), ev.get('peer_interface'), ev.get('peer'))
        hl[ev['accesses']][ev['family']][ev['sock_type']][ev['protocol']][local][peer] = True

    @classmethod
    def from_hashlog(cls, hl):
        for access, family, sock_type, protocol, local_event, peer_event in BaseRule.generate_rules_from_hashlog(hl, 6):
            if access and set(access.split()) & non_peer_accesses:
                peer_event = (None, None, None, None)
            yield cls(access, family, sock_type, local_event, peer_event, log_event=True)


class NetworkRuleset(BaseRuleset):
    """Class to handle and store a collection of network rules"""

    def get_glob(self, path_or_rule):
        """Return the next possible glob. For network rules, that's "network DOMAIN," or "network," (all network)"""
        # XXX return 'network DOMAIN,' if 'network DOMAIN TYPE_OR_PROTOCOL' was given
        return 'network,'


def is_valid_ip(ip):
    if ip == 'none':
        return True
    try:
        ipaddress.ip_address(ip)
        return True
    except ValueError:
        return False



# ----------------------------------------------------------------------
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#    Copyright (C) 2017 Christian Boltz <apparmor@cboltz.de>
#    Copyright (C) 2017 SUSE Linux
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import json
import os
import re
import readline
import subprocess
import sys
from tempfile import NamedTemporaryFile

from apparmor.common import AppArmorException, DebugLogger, readkey
from apparmor.translations import init_translation

_ = init_translation()

# Set up UI logger for separate messages from UI module
debug_logger = DebugLogger('UI')

ARROWS = {'A': 'UP', 'B': 'DOWN', 'C': 'RIGHT', 'D': 'LEFT'}

UI_mode = 'text'
jsonlog = None


def write_json(jsonout):
    jtxt = json.dumps(jsonout, sort_keys=False, separators=(',', ': '))

    if jsonlog:
        jsonlog.write('o ' + jtxt + '\n')

    print(jtxt)
    sys.stdout.flush()


def set_json_mode(cfg):
    """
    Currently this is only used by aa-genprof and aa-logprof, while e.g.
    aa-status generates its own JSON output.

    Remember to bump the JSON API version number if the output commands
    in this file are modified.

    Current known consumers of the JSON output:
    - YaST

    The cfg parameter expects the parsed logprof.conf aka apparmor.aa.cfg.
    """
    global UI_mode, jsonlog

    if int(cfg['settings'].get('json_log', False)):
        jsonlog = NamedTemporaryFile('w', prefix='aa-jsonlog-', delete=False, encoding='utf-8')

    UI_mode = 'json'
    jsonout = {'dialog': 'apparmor-json-version', 'data': '2.12'}
    write_json(jsonout)


def set_text_mode():
    """Output plaintext"""
    global UI_mode
    UI_mode = 'text'


def set_allow_all_mode():
    global UI_mode
    UI_mode = 'allow_all'


# reads the response on command line for json and verifies the response
# for the dialog type
def json_response(dialog_type):
    string = input('\n')

    if jsonlog:
        jsonlog.write('i ' + string + '\n')

    rh = json.loads(string.strip())
    if rh["dialog"] != dialog_type:
        raise AppArmorException('Expected response %s got %s.' % (dialog_type, string))
    return rh


def getkey():
    key = readkey()
    if key == '\x1B':
        key = readkey()
        if key == '[':
            key = readkey()
            if ARROWS.get(key, False):
                key = ARROWS[key]
    return key.strip()


def UI_Info(text):
    """Facility to output normal text"""
    debug_logger.info(text)
    if UI_mode == 'json':
        jsonout = {'dialog': 'info', 'data': text}
        write_json(jsonout)
    else:  # text mode
        sys.stdout.write(text + '\n')


def UI_Important(text):
    """Facility to output important text"""
    debug_logger.debug(text)
    if UI_mode == 'json':
        jsonout = {'dialog': 'important', 'data': text}
        write_json(jsonout)
    else:  # text mode
        sys.stdout.write('\n' + text + '\n')


def get_translated_hotkey(translated, cmsg=''):
    # Originally (\S) was used but with translations it would not work :(
    match = re.search(r'\((\S+)\)', translated)
    if match:
        return match.groups()[0].lower()

    if not cmsg:
        cmsg = 'PromptUser: %s %s' % (_('Invalid hotkey for'), translated)
    raise AppArmorException(cmsg)


def UI_YesNo(text, default):
    debug_logger.debug('UI_YesNo: %s: %s %s', UI_mode, text, default)
    default = default.lower()
    yes = CMDS['CMD_YES']
    no = CMDS['CMD_NO']
    yeskey = get_translated_hotkey(yes)
    nokey = get_translated_hotkey(no)
    ans = 'XXXINVALIDXXX'
    while ans not in ('y', 'n'):
        if UI_mode == 'json':
            jsonout = {'dialog': 'yesno', 'text': text, 'default': default}
            write_json(jsonout)
            hm = json_response('yesno')
            ans = hm['response_key']
        else:  # text mode
            sys.stdout.write('\n' + text + '\n')
            if default == 'y':
                sys.stdout.write('\n[%s] / %s\n' % (yes, no))
            else:
                sys.stdout.write('\n%s / [%s]\n' % (yes, no))
            if UI_mode == 'allow_all':
                ans = nokey
            else:
                ans = getkey()
            if ans:
                # Get back to english from localised answer
                ans = ans.lower()
                if ans == yeskey:
                    ans = 'y'
                elif ans == nokey:
                    ans = 'n'
                elif ans == 'left':
                    default = 'y'
                elif ans == 'right':
                    default = 'n'
                else:
                    ans = 'XXXINVALIDXXX'
                    continue  # If user presses any other button ask again
            else:
                ans = default
    return ans


def UI_YesNoCancel(text, default):
    debug_logger.debug('UI_YesNoCancel: %s: %s %s', UI_mode, text, default)
    default = default.lower()
    yes = CMDS['CMD_YES']
    no = CMDS['CMD_NO']
    cancel = CMDS['CMD_CANCEL']

    yeskey = get_translated_hotkey(yes)
    nokey = get_translated_hotkey(no)
    cancelkey = get_translated_hotkey(cancel)

    ans = 'XXXINVALIDXXX'
    while ans not in ('c', 'n', 'y'):
        if UI_mode == 'json':
            jsonout = {'dialog': 'yesnocancel', 'text': text, 'default': default}
            write_json(jsonout)
            hm = json_response('yesnocancel')
            ans = hm['response_key']
        else:  # text mode
            sys.stdout.write('\n' + text + '\n')
            if default == 'y':
                sys.stdout.write('\n[%s] / %s / %s\n' % (yes, no, cancel))
            elif default == 'n':
                sys.stdout.write('\n%s / [%s] / %s\n' % (yes, no, cancel))
            else:
                sys.stdout.write('\n%s / %s / [%s]\n' % (yes, no, cancel))
            if UI_mode == 'allow_all':
                ans = nokey
            else:
                ans = getkey()
            if ans:
                # Get back to english from localised answer
                ans = ans.lower()
                if ans == yeskey:
                    ans = 'y'
                elif ans == nokey:
                    ans = 'n'
                elif ans == cancelkey:
                    ans = 'c'
                elif ans == 'left':
                    if default == 'n':
                        default = 'y'
                    elif default == 'c':
                        default = 'n'
                elif ans == 'right':
                    if default == 'y':
                        default = 'n'
                    elif default == 'n':
                        default = 'c'
            else:
                ans = default
    return ans


def UI_GetString(text, default):
    debug_logger.debug('UI_GetString: %s: %s %s', UI_mode, text, default)
    string = default
    if UI_mode == 'json':
        jsonout = {'dialog': 'getstring', 'text': text, 'default': default}
        write_json(jsonout)
        string = json_response('getstring')["response"]
    else:  # text mode
        readline.set_startup_hook(lambda: readline.insert_text(default))
        try:
            string = input('\n' + text)
        except EOFError:
            string = ''
        finally:
            readline.set_startup_hook()
    return string.strip()


def UI_GetFile(file):
    debug_logger.debug('UI_GetFile: %s', UI_mode)
    filename = None
    if UI_mode == 'json':
        jsonout = {'dialog': 'getfile', 'text': file['description']}
        write_json(jsonout)
        filename = json_response('getfile')["response"]
    else:  # text mode
        sys.stdout.write(file['description'] + '\n')
        filename = sys.stdin.read()
    return filename


def UI_BusyStart(message):
    debug_logger.debug('UI_BusyStart: %s', UI_mode)
    UI_Info(message)


def UI_BusyStop():
    debug_logger.debug('UI_BusyStop: %s', UI_mode)


def diff(oldprofile, newprofile):
    difftemp = NamedTemporaryFile('w')
    subprocess.call('diff -u -p %s %s > %s' % (oldprofile, newprofile, difftemp.name), shell=True)
    return difftemp


def write_profile_to_tempfile(profile):
    temp = NamedTemporaryFile('w')
    temp.write(profile)
    temp.flush()
    return temp


def generate_diff(oldprofile, newprofile):
    with write_profile_to_tempfile(oldprofile) as oldtemp, \
            write_profile_to_tempfile(newprofile) as newtemp:
        return diff(oldtemp.name, newtemp.name)


def generate_diff_with_comments(oldprofile, newprofile):
    if not os.path.exists(oldprofile):
        raise AppArmorException(_("Can't find existing profile %s to compare changes.") % oldprofile)
    with write_profile_to_tempfile(newprofile) as newtemp:
        return diff(oldprofile, newtemp.name)


def UI_Changes(oldprofile, newprofile, comments=False):
    if not comments:
        difftemp = generate_diff(oldprofile, newprofile)
        header = 'View Changes'
    else:
        difftemp = generate_diff_with_comments(oldprofile, newprofile)
        header = 'View Changes with comments'
    with difftemp:
        UI_ShowFile(header, difftemp.name)


def UI_ShowFile(header, filename):
    if UI_mode == 'json':
        jsonout = {'dialog': 'changes', 'header': header, 'filename': filename}
        write_json(jsonout)
        json_response('changes')["response"]  # wait for response to delay deletion of filename (and ignore response content)
    else:
        subprocess.call(['less', filename])


CMDS = {'CMD_ALLOW': _('(A)llow'),
        'CMD_OTHER': _('(M)ore'),
        'CMD_AUDIT_NEW': _('Audi(t)'),
        'CMD_AUDIT_OFF': _('Audi(t) off'),
        'CMD_AUDIT_FULL': _('Audit (A)ll'),
        # 'CMD_OTHER': '(O)pts',
        'CMD_USER_ON': _('(O)wner permissions on'),
        'CMD_USER_OFF': _('(O)wner permissions off'),
        'CMD_DENY': _('(D)eny'),
        'CMD_ABORT': _('Abo(r)t'),
        'CMD_FINISHED': _('(F)inish'),
        'CMD_ix': _('In(h)erit'),
        'CMD_px': _('(P)rofile'),
        'CMD_px_safe': _('(P)rofile Clean Exec'),
        'CMD_cx': _('(C)hild'),
        'CMD_cx_safe': _('(C)hild Clean Exec'),
        'CMD_nx': _('(N)amed'),
        'CMD_nx_safe': _('(N)amed Clean Exec'),
        'CMD_ux': _('(U)nconfined'),
        'CMD_ux_safe': _('(U)nconfined Clean Exec'),
        'CMD_pix': _('(P)rofile Inherit'),
        'CMD_pix_safe': _('(P)rofile Inherit Clean Exec'),
        'CMD_cix': _('(C)hild Inherit'),
        'CMD_cix_safe': _('(C)hild Inherit Clean Exec'),
        'CMD_nix': _('(N)amed Inherit'),
        'CMD_nix_safe': _('(N)amed Inherit Clean Exec'),
        'CMD_EXEC_IX_ON': _('(X) ix On'),
        'CMD_EXEC_IX_OFF': _('(X) ix Off'),
        'CMD_SAVE': _('(S)ave Changes'),
        'CMD_NEW': _('(N)ew'),
        'CMD_GLOB': _('(G)lob'),
        'CMD_GLOBEXT': _('Glob with (E)xtension'),
        'CMD_ADDHAT': _('(A)dd Requested Hat'),
        'CMD_ADDSUBPROFILE': _('(A)dd Requested Subprofile'),
        'CMD_USEDEFAULT': _('(U)se Default Hat'),
        'CMD_SCAN': _('(S)can system log for AppArmor events'),
        'CMD_HELP': _('(H)elp'),
        'CMD_VIEW_PROFILE': _('(V)iew Profile'),
        'CMD_USE_PROFILE': _('(U)se Profile'),
        'CMD_CREATE_PROFILE': _('(C)reate New Profile'),
        'CMD_UPDATE_PROFILE': _('(U)pdate Profile'),
        'CMD_IGNORE_UPDATE': _('(I)gnore Update'),
        'CMD_SAVE_CHANGES': _('(S)ave Changes'),
        'CMD_SAVE_SELECTED': _('Save Selec(t)ed Profile'),
        'CMD_VIEW_CHANGES': _('(V)iew Changes'),
        'CMD_VIEW_CHANGES_CLEAN': _('View Changes b/w (C)lean profiles'),
        'CMD_VIEW': _('(V)iew'),
        'CMD_ENABLE_REPO': _('(E)nable Repository'),
        'CMD_DISABLE_REPO': _('(D)isable Repository'),
        'CMD_YES': _('(Y)es'),
        'CMD_NO': _('(N)o'),
        'CMD_CANCEL': _('(C)ancel'),
        'CMD_ALL_NET': _('Allow All (N)etwork'),
        'CMD_NET_FAMILY': _('Allow Network Fa(m)ily'),
        'CMD_OVERWRITE': _('(O)verwrite Profile'),
        'CMD_KEEP': _('(K)eep Profile'),
        'CMD_IGNORE_ENTRY': _('(I)gnore')
        }


class PromptQuestion:
    title = None
    headers = None
    explanation = None
    functions = None
    options = None
    default = None
    selected = None
    helptext = None
    already_have_profile = False

    def __init__(self):
        self.headers = []
        self.functions = []
        self.selected = 0

    def promptUser(self, params=''):
        cmd = None
        arg = None
        cmd, arg = self.Text_PromptUser()
        if cmd == 'CMD_ABORT':
            confirm_and_abort()
            cmd = 'XXXINVALIDXXX'
        return (cmd, arg)

    def Text_PromptUser(self):
        title = self.title
        explanation = self.explanation
        headers = self.headers
        functions = self.functions

        default = self.default
        options = self.options
        selected = self.selected
        helptext = self.helptext

        if helptext:
            functions.append('CMD_HELP')

        menu_items = []
        keys = dict()

        for cmd in functions:
            if not CMDS.get(cmd, False):
                raise AppArmorException(_('PromptUser: Unknown command %s') % cmd)

            menutext = CMDS[cmd]

            key = get_translated_hotkey(menutext)
            # Duplicate hotkey
            if keys.get(key, False):
                raise AppArmorException(
                    _('PromptUser: Duplicate hotkey for %(command)s: %(menutext)s ')
                    % {'command': cmd, 'menutext': menutext})

            keys[key] = cmd

            if default and default == cmd:
                menutext = '[%s]' % menutext

            menu_items.append(menutext)

        default_key = 0
        if default and CMDS[default]:
            defaulttext = CMDS[default]
            defmsg = _('PromptUser: Invalid hotkey in default item')

            default_key = get_translated_hotkey(defaulttext, defmsg)

            if not keys.get(default_key, False):
                raise AppArmorException(_('PromptUser: Invalid default %s') % default)

        widest = 0
        header_copy = headers[:]
        while header_copy:
            header = header_copy.pop(0)
            header_copy.pop(0)
            if len(header) > widest:
                widest = len(header)
        widest += 1

        formatstr = '%-' + str(widest) + 's %s\n'

        function_regexp = '^('
        function_regexp += '|'.join(keys.keys())
        if options:
            function_regexp += r'|\d'
        function_regexp += ')$'

        ans = 'XXXINVALIDXXX'
        hdict = dict()
        jsonprompt = {
            'dialog': 'promptuser',
            'title': title,
            'headers': hdict,
            'explanation': explanation,
            'options': options,
            'menu_items': menu_items,
            'default_key': default_key,
        }

        while not re.search(function_regexp, ans, flags=re.IGNORECASE):

            prompt = '\n'
            if title:
                prompt += '= %s =\n\n' % title

            if headers:
                header_copy = headers[:]
                while header_copy:
                    header = header_copy.pop(0)
                    value = header_copy.pop(0)
                    hdict[header] = value
                    prompt += formatstr % (header + ':', value)
                prompt += '\n'

            if explanation:
                prompt += explanation + '\n\n'

            if options:
                for index, option in enumerate(options):
                    if selected == index:
                        format_option = ' [%s - %s]'
                    else:
                        format_option = '  %s - %s '
                    prompt += format_option % (index + 1, option)
                    prompt += '\n'

            prompt += ' / '.join(menu_items)

            if UI_mode == 'json':
                write_json(jsonprompt)
                hm = json_response('promptuser')
                ans = hm["response_key"]
                selected = hm["selected"]

            elif UI_mode == 'allow_all':
                if self.already_have_profile:
                    expected_keys = ['CMD_px', 'CMD_ix', 'CMD_ALLOW', 'CMD_SAVE_CHANGES']
                else:
                    expected_keys = ['CMD_ix', 'CMD_ALLOW', 'CMD_SAVE_CHANGES']
                for exp in expected_keys:
                    if exp in functions:
                        ans = get_translated_hotkey(CMDS[exp])
                        break

            else:  # text mode
                sys.stdout.write(prompt + '\n')
                ans = getkey().lower()

            if ans:
                if ans == 'up':
                    if options and selected > 0:
                        selected -= 1
                    ans = 'XXXINVALIDXXX'

                elif ans == 'down':
                    if options and selected < len(options) - 1:
                        selected += 1
                    ans = 'XXXINVALIDXXX'

    #             elif keys.get(ans, False) == 'CMD_HELP':
    #                 sys.stdout.write('\n%s\n' %helptext)
    #                 ans = 'XXXINVALIDXXX'

                elif is_number(ans) == 10:
                    # If they hit return choose default option
                    ans = default_key

                elif options and re.search(r'^\d$', ans):
                    ans = int(ans)
                    if ans > 0 and ans <= len(options):
                        selected = ans - 1
                    ans = 'XXXINVALIDXXX'

            if keys.get(ans, False) == 'CMD_HELP' and UI_mode != 'json':
                sys.stdout.write('\n%s\n' % helptext)
                ans = 'again'

        if keys.get(ans, False):
            ans = keys[ans]

        return ans, selected


def confirm_and_abort():
    ans = UI_YesNo(_('Are you sure you want to abandon this set of profile changes and exit?'), 'n')
    if ans == 'y':
        UI_Info(_('Abandoning all changes.'))
        sys.exit(0)


def is_number(number):
    try:
        return int(number)
    except (TypeError, ValueError):
        return False



# ----------------------------------------------------------------------
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#    Copyright (C) 2015 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import re

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.regex import RE_PROFILE_RLIMIT, strip_quotes
from apparmor.rule import BaseRule, BaseRuleset, parse_comment, quote_if_needed
from apparmor.translations import init_translation

_ = init_translation()

rlimit_size = ['fsize', 'data', 'stack', 'core', 'rss', 'as', 'memlock', 'msgqueue']  # NUMBER ( 'K' | 'M' | 'G' )
rlimit_number = ['ofile', 'nofile', 'locks', 'sigpending', 'nproc', 'rtprio']
rlimit_time = ['cpu', 'rttime']  # number + time unit (cpu in seconds+, rttime in us+)
rlimit_nice = ['nice']  # a number between -20 and 19.

rlimit_all = rlimit_size + rlimit_number + rlimit_time + rlimit_nice

RE_NUMBER_UNIT = re.compile(r'^(?P<number>[0-9]+)\s*(?P<unit>[a-zA-Z]*)$')
RE_NUMBER = re.compile('^[0-9]+$')
RE_UNIT_SIZE = re.compile(r'^[0-9]+\s*([KMG]B?)?$')
RE_NICE = re.compile('^(-20|-[01]?[0-9]|[01]?[0-9])$')


class RlimitRule(BaseRule):
    """Class to handle and store a single rlimit rule"""

    # Nothing external should reference this class, all external users
    # should reference the class field RlimitRule.ALL
    class __RlimitAll:
        pass

    ALL = __RlimitAll

    rule_name = 'rlimit'
    _match_re = RE_PROFILE_RLIMIT

    def __init__(self, rlimit, value, audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny, allow_keyword=allow_keyword,
                         comment=comment, log_event=log_event, priority=priority)

        # rlimit rules don't support priority, allow keyword, audit or deny
        self.ensure_modifiers_not_supported()

        if isinstance(rlimit, str):
            if rlimit in rlimit_all:
                self.rlimit = rlimit
            else:
                raise AppArmorException('Unknown rlimit keyword in rlimit rule: %s' % rlimit)
        else:
            raise AppArmorBug('Passed unknown object to %s: %s' % (type(self).__name__, str(rlimit)))

        self.value = None
        self.value_as_int = None
        self.all_values = False
        if value == self.ALL:
            self.all_values = True
        elif isinstance(value, str):
            if not value.strip():
                raise AppArmorBug('Empty value in rlimit rule')

            elif rlimit in rlimit_size:
                if not RE_UNIT_SIZE.match(value):
                    raise AppArmorException('Invalid value or unit in rlimit %s %s rule' % (rlimit, value))
                self.value_as_int = self.size_to_int(value)

            elif rlimit in rlimit_number:
                if not RE_NUMBER.match(value):
                    raise AppArmorException('Invalid value in rlimit %s %s rule' % (rlimit, value))
                self.value_as_int = int(value)

            elif rlimit in rlimit_time:
                if not RE_NUMBER_UNIT.match(value):
                    raise AppArmorException('Invalid value in rlimit %s %s rule' % (rlimit, value))
                number, unit = split_unit(value)

                if rlimit == 'rttime':
                    self.value_as_int = self.time_to_int(value, 'us')
                else:
                    self.value_as_int = self.time_to_int(value, 'seconds')

            elif rlimit in rlimit_nice:  # pragma: no branch - "if rlimit in rlimit_all:" above avoids the need for an "else:" branch
                if not RE_NICE.match(value):
                    raise AppArmorException('Invalid value or unit in rlimit %s %s rule' % (rlimit, value))
                self.value_as_int = 0 - int(value)  # lower numbers mean a higher limit for nice

            # still here? fine :-)
            self.value = value
        else:
            raise AppArmorBug('Passed unknown object to %s: %s' % (type(self).__name__, str(value)))

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        """parse raw_rule and return instance of this class"""

        comment = parse_comment(matches)

        if matches.group('rlimit'):
            rlimit = strip_quotes(matches.group('rlimit'))
        else:
            raise AppArmorException(_("Invalid rlimit rule '%s' - keyword missing") % raw_rule)  # pragma: no cover - would need breaking the regex

        if matches.group('value'):
            if matches.group('value') == 'infinity':
                value = cls.ALL
            else:
                value = strip_quotes(matches.group('value'))
        else:
            raise AppArmorException(_("Invalid rlimit rule '%s' - value missing") % raw_rule)  # pragma: no cover - would need breaking the regex

        return cls(rlimit, value, comment=comment)

    def get_clean(self, depth=0):
        """return rule (in clean/default formatting)"""

        space = '  ' * depth

        if self.rlimit:
            rlimit = ' %s' % quote_if_needed(self.rlimit)
        else:
            raise AppArmorBug('Empty rlimit in rlimit rule')

        if self.all_values:
            value = ' <= infinity'
        elif self.value:
            value = ' <= %s' % quote_if_needed(self.value)
        else:
            raise AppArmorBug('Empty value in rlimit rule')

        return ('%s%sset rlimit%s%s,%s' % (space, self.modifiers_str(), rlimit, value, self.comment))

    def size_to_int(self, value):
        number, unit = split_unit(value)

        if not unit:
            pass
        elif unit == 'K' or unit == 'KB':
            number = number * 1024
        elif unit == 'M' or unit == 'MB':
            number = number * 1024 * 1024
        elif unit == 'G' or unit == 'GB':
            number = number * 1024 * 1024 * 1024
        else:
            raise AppArmorException('Unknown unit %s in rlimit %s %s' % (unit, self.rlimit, value))

        return number

    def time_to_int(self, value, default_unit):
        number, unit = split_unit(value)

        if not unit:
            unit = default_unit

        if unit in ('us', 'microsecond', 'microseconds'):
            number = number / 1000000.0
            if default_unit == 'seconds':
                raise AppArmorException(_('Invalid unit in rlimit cpu %s rule') % value)
        elif unit in ('ms', 'millisecond', 'milliseconds'):
            number = number / 1000.0
            if default_unit == 'seconds':
                raise AppArmorException(_('Invalid unit in rlimit cpu %s rule') % value)
        elif unit in ('s', 'sec', 'second', 'seconds'):  # manpage doesn't list sec
            pass
        elif unit in ('min', 'minute', 'minutes'):
            number = number * 60
        elif unit in ('h', 'hour', 'hours'):
            number = number * 60 * 60
        elif unit in ('d', 'day', 'days'):  # manpage doesn't list 'd'
            number = number * 60 * 60 * 24
        elif unit in ('week', 'weeks'):
            number = number * 60 * 60 * 24 * 7
        else:
            raise AppArmorException('Unknown unit %s in rlimit %s %s' % (unit, self.rlimit, value))

        return number

    def _is_covered_localvars(self, other_rule):
        """check if other_rule is covered by this rule object"""

        if not self._is_covered_plain(self.rlimit, False, other_rule.rlimit, False, 'rlimit'):  # rlimit can't be ALL, therefore using False
            return False

        if not other_rule.value and not other_rule.all_values:
            raise AppArmorBug('No target profile specified in other rlimit rule')

        if not self.all_values:
            if other_rule.all_values:
                return False
            if other_rule.value_as_int > self.value_as_int:
                return False

        # still here? -> then it is covered
        return True

    def _is_equal_localvars(self, other_rule, strict):
        """compare if rule-specific variables are equal"""

        if self.rlimit != other_rule.rlimit:
            return False

        if (self.value_as_int != other_rule.value_as_int
                or self.all_values != other_rule.all_values):
            return False

        return True

    def _logprof_header_localvars(self):
        if self.all_values:
            values_txt = 'infinity'
        else:
            values_txt = self.value

        return (
            _('Rlimit'), self.rlimit,
            _('Value'),  values_txt,
        )


class RlimitRuleset(BaseRuleset):
    """Class to handle and store a collection of rlimit rules"""

    def get_glob(self, path_or_rule):
        """Return the next possible glob. For rlimit rules, that can mean changing the value to 'infinity'"""
        # XXX implement all options mentioned above ;-)
        raise AppArmorBug('get_glob() is not (yet) available for this rule type')


def split_unit(value):
    matches = RE_NUMBER_UNIT.match(value)
    if not matches:
        raise AppArmorBug("Invalid value given to split_unit: %s" % value)

    number = int(matches.group('number'))
    unit = matches.group('unit') or ''

    return number, unit



# ----------------------------------------------------------------------
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#    Copyright (C) 2014 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

from abc import ABCMeta, abstractmethod

from apparmor.aare import AARE
from apparmor.common import AppArmorBug, AppArmorException, hasher
from apparmor.regex import strip_quotes
from apparmor.translations import init_translation

_ = init_translation()


class BaseRule(metaclass=ABCMeta):
    """Base class to handle and store a single rule"""

    # decides if the (G)lob and Glob w/ (E)xt options are displayed
    can_glob = False
    can_glob_ext = False

    # defines if the (N)ew option is displayed
    can_edit = False

    # defines if the '(O)wner permissions on/off' option is displayed
    can_owner = False

    rule_name = 'base'

    _match_re = None

    def __init__(self, audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):
        """initialize variables needed by all rule types"""

        self._store_priority(priority)
        self.audit = audit
        self.deny = deny
        self.allow_keyword = allow_keyword
        self.comment = comment
        self.log_event = log_event

        # Set only in the parse() class method
        self.raw_rule = None

    def _store_priority(self, priority):
        if priority is None:  # default priority
            self.priority = None
            return

        try:
            ipriority = int(priority)
        except ValueError:
            raise AppArmorException("Invalid value for priority '%s'" % priority)

        if ipriority < -1000 or ipriority > 1000:
            raise AppArmorException('priority %d out of range' % (ipriority))

        self.priority = ipriority

    def _aare_or_all(self, rulepart, partname, is_path, log_event, empty_ok=False):
        """checks rulepart and returns
           - (AARE, False) if rulepart is a (non-empty) string
           - (None, True) if rulepart is all_obj (typically *Rule.ALL)
           - raises AppArmorBug if rulepart is an empty string or has a wrong type

           Parameters:
           - rulepart: the rule part to check (string or *Rule.ALL object)
           - partname: the name of the rulepart (for example 'peer', used for exception messages)
           - is_path (passed through to AARE)
           - log_event (passed through to AARE)
           """

        if rulepart == self.ALL:
            return None, True
        elif isinstance(rulepart, str):
            if not rulepart.strip() and not empty_ok:
                raise AppArmorBug(
                    'Passed empty %(partname)s to %(classname)s: %(rulepart)s'
                    % {'partname': partname, 'classname': self.__class__.__name__, 'rulepart': str(rulepart)})
            return AARE(rulepart, is_path=is_path, log_event=log_event), False
        else:
            raise AppArmorBug(
                'Passed unknown %(partname)s to %(classname)s: %(rulepart)s'
                % {'partname': partname, 'classname': self.__class__.__name__, 'rulepart': str(rulepart)})

    def __repr__(self):
        return '<%s> %s' % (self.__class__.__name__, self.get_raw())

    @classmethod
    def match(cls, raw_rule):
        """return True if raw_rule matches the class (main) regex, False otherwise
           Note: This function just provides an answer to "is this your job?".
                 It does not guarantee that the rule is completely valid."""
        return bool(cls._match(raw_rule))

    @classmethod
    def _match(cls, raw_rule):
        """parse raw_rule and return regex match object"""
        if cls._match_re is None:
            if cls is BaseRule:
                raise AppArmorBug("BaseRule class methods should not be called directly.")
            else:
                raise NotImplementedError("'%s' needs to implement _match(), but didn't" % (str(cls)))
        return cls._match_re.search(raw_rule)

    @classmethod
    def create_instance(cls, raw_rule):
        """parse raw_rule and return instance of this class"""
        matches = cls._match(raw_rule)
        if not matches:
            raise AppArmorException(_("Invalid %s rule '%s'") % (cls.rule_name, raw_rule))

        rule = cls._create_instance(raw_rule, matches)
        rule.raw_rule = raw_rule.strip()
        return rule

    @classmethod
    @abstractmethod
    def _create_instance(cls, raw_rule, matches):
        """returns a Rule object created from parsing the raw rule.
           required to be implemented by subclasses; raise exception if not"""
        raise NotImplementedError("'%s' needs to implement _create_instance(), but didn't" % (str(cls)))

    @staticmethod
    def generate_rules_from_hashlog(hashlog, nb_keys):
        """yields all key sequences from a hashlog of depth nb_keys"""
        stack = [(hashlog, [], nb_keys)]

        while stack:
            items, path, depth = stack.pop()

            if depth == 0:
                yield path
                continue

            for next_key in items:
                stack.append((items[next_key], path + [next_key], depth - 1))

    @classmethod
    def create_from_ev(cls, ev):
        """returns a rule that would allow an event, or None if the event is invalid"""
        hl = hasher()
        cls.hashlog_from_event(hl, ev)
        return next(cls.from_hashlog(hl), None)

    @staticmethod
    def hashlog_from_event(hl, ev):
        """stores an event in the hashlog"""
        raise NotImplementedError('hashlog_from_event should be called on a rule class and not directly on BaseRule.')

    @classmethod
    def from_hashlog(cls, hl):
        """constructs and yields all rules that would allow denials stored in a hashlog"""
        raise NotImplementedError("'%s' needs to implement from_hashlog(), but didn't" % (str(cls)))

    @abstractmethod
    def get_clean(self, depth=0):
        """return clean rule (with default formatting, and leading whitespace as specified in the depth parameter)"""

    def get_raw(self, depth=0):
        """return raw rule (with original formatting, and leading whitespace in the depth parameter)"""
        if self.raw_rule:
            return '%s%s' % ('  ' * depth, self.raw_rule)
        else:
            return self.get_clean(depth)

    def is_covered(self, other_rule, check_allow_deny=True, check_audit=False, check_priority=True):
        """check if other_rule is covered by this rule object"""

        if type(other_rule) is not type(self):
            raise AppArmorBug('Passes %s instead of %s' % (str(other_rule), self.__class__.__name__))

        if check_allow_deny and self.deny != other_rule.deny:
            return False

        if other_rule.deny and not self.deny:
            return False

        if check_audit and other_rule.audit != self.audit:
            return False

        if other_rule.audit and not self.audit:
            return False

        if check_priority and (self.priority or 0) < (other_rule.priority or 0):
            return False

        # still here? -> then the common part is covered, check rule-specific things now
        return self._is_covered_localvars(other_rule)

    @abstractmethod
    def _is_covered_localvars(self, other_rule):
        """check if the rule-specific parts of other_rule is covered by this rule object"""

    def _is_covered_plain(self, self_value, self_all, other_value, other_all, cond_name):
        """check if other_* is covered by self_* - for plain str, int etc."""

        if not other_value and not other_all:
            raise AppArmorBug('No %(cond_name)s specified in other %(rule_name)s rule' % {'cond_name': cond_name, 'rule_name': self.rule_name})

        if not self_all:
            if other_all:
                return False
            if self_value != other_value:
                return False

        # still here? -> then it is covered
        return True

    def _is_covered_list(self, self_value, self_all, other_value, other_all, cond_name, sanity_check=True):
        """check if other_* is covered by self_* - for lists"""

        if sanity_check and not other_value and not other_all:
            raise AppArmorBug('No %(cond_name)s specified in other %(rule_name)s rule' % {'cond_name': cond_name, 'rule_name': self.rule_name})

        if not self_all:
            if other_all:
                return False
            if not other_value.issubset(self_value):
                return False

        # still here? -> then it is covered
        return True

    def _is_covered_aare(self, self_value, self_all, other_value, other_all, cond_name):
        """check if other_* is covered by self_* - for AARE"""

        if not other_value and not other_all:
            raise AppArmorBug('No %(cond_name)s specified in other %(rule_name)s rule' % {'cond_name': cond_name, 'rule_name': self.rule_name})

        if not self_all:
            if other_all:
                return False
            if not self_value.match(other_value):
                return False

        # still here? -> then it is covered
        return True

    def is_equal(self, other_rule, strict=False):
        """compare if other_rule == self
           Calls _is_equal_localvars() to compare rule-specific variables"""

        if ((self.priority or 0) != (other_rule.priority or 0)
                or self.audit != other_rule.audit
                or self.deny != other_rule.deny):
            return False

        if strict and (
            self.priority != other_rule.priority
            or self.allow_keyword != other_rule.allow_keyword
            or self.comment != other_rule.comment
            or self.raw_rule != other_rule.raw_rule
        ):
            return False

        if type(other_rule) is not type(self):
            raise AppArmorBug('Passed non-{} rule: {}'.format(self.rule_name, other_rule))

        return self._is_equal_localvars(other_rule, strict)

    def _is_equal_aare(self, self_value, self_all, other_value, other_all, cond_name):
        """check if other_* is the same as self_* - for AARE"""

        if not other_value and not other_all:
            raise AppArmorBug('No %(cond_name)s specified in other %(rule_name)s rule' % {'cond_name': cond_name, 'rule_name': self.rule_name})

        if self_all != other_all:
            return False

        if self_value and not self_value.is_equal(other_value):
            return False

        # still here? -> then it is equal
        return True

    @abstractmethod
    def _is_equal_localvars(self, other_rule, strict):
        """compare if rule-specific variables are equal"""

    def severity(self, sev_db):
        """return severity of this rule, which can be:
           - a number between 0 and 10, where 0 means harmless and 10 means critical,
           - "unknown" (to be exact: the value specified for "unknown" as set when loading the severity database), or
           - sev_db.NOT_IMPLEMENTED if no severity check is implemented for this rule type.
           sev_db must be an apparmor.severity.Severity object."""
        return sev_db.NOT_IMPLEMENTED

    def logprof_header(self):
        """return the headers (human-readable version of the rule) to display in aa-logprof for this rule object
           returns {'label1': 'value1', 'label2': 'value2'}"""

        headers = []
        qualifier = []

        if self.priority:
            qualifier.append('priority=%s' % self.priority)

        if self.audit:
            qualifier.append('audit')

        if self.deny:
            qualifier.append('deny')
        elif self.allow_keyword:
            qualifier.append('allow')

        if qualifier:
            headers.extend((_('Qualifier'), ' '.join(qualifier)))

        headers.extend(self._logprof_header_localvars())

        return headers

    @abstractmethod
    def _logprof_header_localvars(self):
        """return the headers (human-readable version of the rule) to display in aa-logprof for this rule object
           returns {'label1': 'value1', 'label2': 'value2'}"""

    # NOTE: edit_header, validate_edit, and store_edit are not implemented by every subclass.
    def edit_header(self):
        """return the prompt for, and the path to edit when using '(N)ew'"""
        raise NotImplementedError("'%s' needs to implement edit_header(), but didn't" % (str(self)))

    def validate_edit(self, newpath):
        """validate the new path.
           Returns True if it covers the previous path, False if it doesn't."""
        raise NotImplementedError("'%s' needs to implement validate_edit(), but didn't" % (str(self)))

    def store_edit(self, newpath):
        """store the changed path.
           This is done even if the new path doesn't match the original one."""
        raise NotImplementedError("'%s' needs to implement store_edit(), but didn't" % (str(self)))

    def modifiers_str(self):
        """return priority, allow/deny, and audit keyword as string, including whitespace"""

        if self.priority is not None:
            prioritystr = 'priority=%s ' % self.priority
        else:
            prioritystr = ''

        if self.audit:
            auditstr = 'audit '
        else:
            auditstr = ''

        if self.deny:
            allowstr = 'deny '
        elif self.allow_keyword:
            allowstr = 'allow '
        else:
            allowstr = ''

        return '%s%s%s' % (prioritystr, auditstr, allowstr)

    def ensure_modifiers_not_supported(self):
        if self.audit:
            raise AppArmorBug('Attempt to initialize %s with audit flag' % self.__class__.__name__)
        if self.deny:
            raise AppArmorBug('Attempt to initialize %s with deny flag' % self.__class__.__name__)
        if self.allow_keyword:
            raise AppArmorBug('Attempt to initialize %s with allow keyword' % self.__class__.__name__)
        if self.priority is not None:
            raise AppArmorBug('Attempt to initialize %s with priority' % self.__class__.__name__)


class BaseRuleset:
    """Base class to handle and store a collection of rules"""

    # decides if the (G)lob and Glob w/ (E)xt options are displayed
    # XXX TODO: remove in all *Ruleset classes (moved to *Rule)
    can_glob = True
    can_glob_ext = False

    def __init__(self):
        """initialize variables needed by all ruleset types
           Do not override in child class unless really needed - override _init_vars() instead"""
        self.rules = []
        self._init_vars()

    def _init_vars(self):
        """called by __init__() and delete_all_rules() - override in child class to initialize more variables"""

    def __repr__(self):
        classname = self.__class__.__name__
        if self.rules:
            return '<%s>\n' % classname + '\n'.join(self.get_raw(1)) + '</%s>' % classname
        else:
            return '<%s (empty) />' % classname

    def add(self, rule, cleanup=False):
        """add a rule object
           if cleanup is specified, delete rules that are covered by the new rule
           (the difference to delete_duplicates() is: cleanup only deletes rules that
           are covered by the new rule, but keeps other, unrelated superfluous rules)
        """
        deleted = 0

        if cleanup:
            oldrules = self.rules
            self.rules = []

            for oldrule in oldrules:
                if not rule.is_covered(oldrule):
                    self.rules.append(oldrule)
                else:
                    deleted += 1

        self.rules.append(rule)

        return deleted

    def get_raw(self, depth=0):
        """return all raw rules (if possible/not modified in their original formatting).
           Returns an array of lines, with depth * leading whitespace"""

        data = []
        for rule in self.rules:
            data.append(rule.get_raw(depth))

        if data:
            data.append('')

        return data

    def get_clean(self, depth=0):
        """return all rules (in clean/default formatting)
           Returns an array of lines, with depth * leading whitespace"""

        allow_rules = []
        deny_rules = []

        for rule in self.rules:
            if rule.deny:
                deny_rules.append(rule.get_clean(depth))
            else:
                allow_rules.append(rule.get_clean(depth))

        allow_rules.sort()
        deny_rules.sort()

        cleandata = []

        if deny_rules:
            cleandata += deny_rules
            cleandata.append('')

        if allow_rules:
            cleandata += allow_rules
            cleandata.append('')

        return cleandata

    def get_clean_unsorted(self, depth=0):
        """return all rules (in clean/default formatting) in original order
           Returns an array of lines, with depth * leading whitespace"""

        all_rules = []

        for rule in self.rules:
            all_rules.append(rule.get_clean(depth))

        if all_rules:
            all_rules.append('')

        return all_rules

    def is_covered(self, rule, check_allow_deny=True, check_audit=False):
        """return True if rule is covered by existing rules, otherwise False"""

        for r in self.rules:
            if r.is_covered(rule, check_allow_deny, check_audit):
                return True

        return False

#    def is_log_covered(self, parsed_log_event, check_allow_deny=True, check_audit=False):
#        """return True if parsed_log_event is covered by existing rules, otherwise False"""
#
#        rule_obj = self.new_rule()
#        rule_obj.set_log(parsed_log_event)
#
#        return self.is_covered(rule_obj, check_allow_deny, check_audit)

    def delete(self, rule):
        """Delete rule from rules"""

        rule_to_delete = False
        i = 0
        for r in self.rules:
            if r.is_equal(rule):
                rule_to_delete = True
                break
            i = i + 1

        if rule_to_delete:
            self.rules.pop(i)
        else:
            raise AppArmorBug('Attempt to delete non-existing rule %s' % rule.get_raw(0))

    def delete_duplicates(self, include_rules):
        """Delete duplicate rules.
           include_rules must be a *_rules object or None"""

        deleted = 0

        # delete rules that are covered by include files
        if include_rules:
            oldrules = self.rules
            self.rules = []
            for rule in oldrules:
                if include_rules.is_covered(rule, True, False):
                    deleted += 1
                else:
                    self.rules.append(rule)

        # de-duplicate rules inside the profile
        deleted += self.delete_in_profile_duplicates()
        self.rules.reverse()
        deleted += self.delete_in_profile_duplicates()  # search again in reverse order - this will find the remaining duplicates
        self.rules.reverse()  # restore original order for raw output

        return deleted

    def delete_in_profile_duplicates(self):
        """Delete duplicate rules inside a profile"""

        deleted = 0
        oldrules = self.rules
        self.rules = []

        for rule in oldrules:
            if not self.is_covered(rule, True, False):
                self.rules.append(rule)
            else:
                deleted += 1

        return deleted

    def get_glob_ext(self, path_or_rule):
        """returns the next possible glob with extension (for file rules only).
           For all other rule types, raise an exception"""
        raise NotImplementedError("get_glob_ext is not available for this rule type!")


def check_and_split_list(lst, allowed_keywords, all_obj, classname, keyword_name, allow_empty_list=False):
    """check if lst is all_obj or contains only items listed in allowed_keywords"""

    if lst == all_obj:
        return None, True, None
    elif isinstance(lst, str):
        result_list = {lst}
    elif isinstance(lst, (list, tuple, set)) and (lst or allow_empty_list):
        result_list = set(lst)
    else:
        raise AppArmorBug(
            'Passed unknown %(type)s object to %(classname)s: %(unknown_object)s'
            % {'type': type(lst), 'classname': classname, 'unknown_object': str(lst)})

    unknown_items = set()
    for item in result_list:
        if not item.strip():
            raise AppArmorBug(
                'Passed empty %(keyword_name)s to %(classname)s'
                % {'keyword_name': keyword_name, 'classname': classname})
        if item not in allowed_keywords:
            unknown_items.add(item)

    return result_list, False, unknown_items


def logprof_value_or_all(value, all_values):
    """helper for logprof_header() to return 'all' (if all_values is True) or the specified value.
       For some types, the value is made more readable."""

    if all_values:
        return _('ALL')

    if isinstance(value, AARE):
        return value.regex
    elif isinstance(value, (set, list, tuple)):
        return ' '.join(sorted(value))
    else:
        return value


def parse_comment(matches):
    """returns the comment (with a leading space) from the matches object"""
    comment = ''
    if matches.group('comment'):
        # include a space so that we don't need to add it everywhere when writing the rule
        comment = ' %s' % matches.group('comment')
    return comment


def parse_modifiers(matches):
    """returns priority, audit, deny, allow_keyword and comment from the
    matches object
       - priority is a number or None
       - audit, deny and allow_keyword are True/False
       - comment is the comment with a leading space"""

    priority = None
    if matches.group('priority'):
        priority = int(matches.group('priority'))

    audit = False
    if matches.group('audit'):
        audit = True

    deny = False
    allow_keyword = False

    allowstr = matches.group('allow')
    if allowstr:
        if allowstr.strip() == 'allow':
            allow_keyword = True
        elif allowstr.strip() == 'deny':
            deny = True
        else:
            raise AppArmorBug("Invalid allow/deny keyword %s" % allowstr)

    comment = parse_comment(matches)

    return (priority, audit, deny, allow_keyword, comment)


def quote_if_needed(data):
    """quote data if it contains whitespace"""
    if ' ' in data:
        data = '"' + data + '"'
    return data


def check_dict_keys(d, possible_keys, type_all):
    """Check that all keys in dictionary are among possible keys"""
    if d == type_all or d == {}:
        return type_all
    if not possible_keys >= d.keys():
        raise AppArmorException(f'Incorrect key in dict {d}. Possible keys are {possible_keys},')
    return d


def initialize_cond_dict(d, keys, suffix, type_all):
    out = {
        key: strip_quotes(d[f'{key}{suffix}'])
        for key in keys
        if f'{key}{suffix}' in d and d[f'{key}{suffix}'] is not None
    }
    return out if out != {} else type_all


def tuple_to_dict(t, keys):
    d = {}
    for idx, k in enumerate(keys):
        if t[idx] is not None:
            d[k] = t[idx]
    return d


def print_dict_values(d, type_all, prefix=None):
    if d == type_all:
        return ''
    to_print = ' '.join(f'{k}={quote_if_needed(str(v))}' for k, v in d.items())
    if prefix:
        return f' {prefix}=({to_print})'
    else:
        return f' {to_print}'



# ----------------------------------------------------------------------
#    Copyright (C) 2015 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import re

from apparmor.common import convert_regexp, AppArmorBug, AppArmorException


path_entry = r"(/[^\n,]*|(@{[^\n,}]*}[^\n,]*))"
# XXX Matching using this regex is not perfect but should be enough for practical scenarios. Limitations are:
#  - We do not look recursively '{', therefore some weird-but-valid string such as {{/foo,/bar},{/baz,/qux}} are denied
#  - Variables are not replaced: we have to accept "@{" regardless of whether it contains an actual path
AARE_PATH_REGEX = re.compile(
    r"^" + path_entry
    + r"|{" + path_entry + "," + path_entry + "*}$"
)


class AARE:
    """AARE (AppArmor Regular Expression) wrapper class"""

    def __init__(self, regex, is_path, log_event=None):
        """Initialize instance for the given AppArmor regex.
        If is_path is true, the regex is expected to be a path and therefore must start with / or a variable."""
        # using the specified variables when matching.

        if is_path and not AARE_PATH_REGEX.match(regex):
            raise AppArmorException("Path doesn't start with / or variable: %s" % regex)

        if log_event:
            self.orig_regex = regex
            self.regex = convert_expression_to_aare(regex)
        else:
            self.orig_regex = None
            self.regex = regex

        self._regex_compiled = None  # done on first use in match() - that saves us some re.compile() calls
        # self.variables = variables  # XXX

    def __repr__(self):
        """returns a "printable" representation of object"""
        return type(self).__name__ + "('%s')" % self.regex

    def __eq__(self, other):
        """check if the given object is equal
           Note that the == check is more strict than is_equal() - it doesn't accept if other is a string instead of AARE"""

        if isinstance(other, type(self)):
            return self.regex == other.regex

        return False

    def __deepcopy__(self, memo):
        # thanks to http://bugs.python.org/issue10076, we need to implement this ourself
        if self.orig_regex:
            return type(self)(self.orig_regex, is_path=False, log_event=True)
        else:
            return type(self)(self.regex, is_path=False)

    # check if a regex is a plain path (not containing variables, alternations or wildcards)
    # some special characters are probably not covered by the plain_path regex (if in doubt, better error out on the safe side)
    plain_path = re.compile('^[0-9a-zA-Z/._-]+$')

    def match(self, expression):
        """check if the given expression (string or instance of this class) matches the regex"""

        if isinstance(expression, type(self)):
            if expression.orig_regex:
                expression = expression.orig_regex
            elif self.plain_path.match(expression.regex):
                # regex doesn't contain variables or wildcards, therefore handle it as plain path
                expression = expression.regex
            else:
                return self.is_equal(expression)  # better safe than sorry
        elif not isinstance(expression, str):
            raise AppArmorBug('match() called with unknown object: %s' % str(expression))

        if self._regex_compiled is None:
            self._regex_compiled = re.compile(convert_regexp(self.regex))

        return bool(self._regex_compiled.match(expression))

    def is_equal(self, expression):
        """check if the given expression is equal"""

        if isinstance(expression, type(self)):
            return self.regex == expression.regex
        elif isinstance(expression, str):
            return self.regex == expression
        else:
            raise AppArmorBug('is_equal() called with unknown object: %s' % str(expression))

    def glob_path(self):
        """Glob the given file or directory path"""
        if self.regex.endswith('/'):
            if self.regex.endswith(('/**/', '/*/')):
                # /foo/**/ and /foo/*/ => /**/
                newpath = re.sub(r'/[^/]+/\*{1,2}/$', '/**/', self.regex)
            elif re.search(r'/[^/]+\*\*[^/]*/$', self.regex):
                # /foo**/ and /foo**bar/ => /**/
                newpath = re.sub(r'/[^/]+\*\*[^/]*/$', '/**/', self.regex)
            elif re.search(r'/\*\*[^/]+/$', self.regex):
                # /**bar/ => /**/
                newpath = re.sub(r'/\*\*[^/]+/$', '/**/', self.regex)
            else:
                newpath = re.sub('/[^/]+/$', '/*/', self.regex)
        else:
            if self.regex.endswith(('/**', '/*')):
                # /foo/** and /foo/* => /**
                newpath = re.sub(r'/[^/]+/\*{1,2}$', '/**', self.regex)
            elif re.search(r'/[^/]*\*\*[^/]+$', self.regex):
                # /**foo and /foor**bar => /**
                newpath = re.sub(r'/[^/]*\*\*[^/]+$', '/**', self.regex)
            elif re.search(r'/[^/]+\*\*$', self.regex):
                # /foo** => /**
                newpath = re.sub(r'/[^/]+\*\*$', '/**', self.regex)
            else:
                newpath = re.sub('/[^/]+$', '/*', self.regex)
        return type(self)(newpath, False)

    def glob_path_withext(self):
        """Glob given file path with extension
           Files without extensions and directories won't be changed"""
        # match /**.ext and /*.ext
        match = re.search(r'/\*{1,2}(\.[^/]+)$', self.regex)
        if match:
            # /foo/**.ext and /foo/*.ext => /**.ext
            newpath = re.sub(r'/[^/]+/\*{1,2}\.[^/]+$', '/**' + match.groups()[0], self.regex)
        elif re.search(r'/[^/]+\*\*[^/]*\.[^/]+$', self.regex):
            # /foo**.ext and /foo**bar.ext => /**.ext
            match = re.search(r'/[^/]+\*\*[^/]*(\.[^/]+)$', self.regex)
            newpath = re.sub(r'/[^/]+\*\*[^/]*\.[^/]+$', '/**' + match.groups()[0], self.regex)
        elif re.search(r'/\*\*[^/]+\.[^/]+$', self.regex):
            # /**foo.ext => /**.ext
            match = re.search(r'/\*\*[^/]+(\.[^/]+)$', self.regex)
            newpath = re.sub(r'/\*\*[^/]+\.[^/]+$', '/**' + match.groups()[0], self.regex)
        else:
            newpath = self.regex
            match = re.search(r'(\.[^/]+)$', self.regex)
            if match:
                newpath = re.sub(r'/[^/]+(\.[^/]+)$', '/*' + match.groups()[0], self.regex)
        return type(self)(newpath, False)


def convert_expression_to_aare(expression):
    """convert an expression (taken from audit.log) to an AARE string"""

    aare_escape_chars = ['\\', '?', '*', '[', ']', '{', '}', '"', '!']
    for char in aare_escape_chars:
        expression = expression.replace(char, '\\' + char)

    return expression



# ----------------------------------------------------------------------
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#    Copyright (C) 2020 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import re

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.regex import RE_PROFILE_VARIABLE, strip_quotes
from apparmor.rule import BaseRule, BaseRuleset, parse_comment, quote_if_needed
from apparmor.translations import init_translation

_ = init_translation()


class VariableRule(BaseRule):
    """Class to handle and store a single variable rule"""

    rule_name = 'variable'
    _match_re = RE_PROFILE_VARIABLE

    def __init__(self, varname, mode, values, audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny, allow_keyword=allow_keyword,
                         comment=comment, log_event=log_event, priority=priority)

        # variables don't support priority, allow keyword, audit or deny
        self.ensure_modifiers_not_supported()

        if not isinstance(varname, str):
            raise AppArmorBug('Passed unknown type for varname to %s: %s' % (self.__class__.__name__, varname))
        if not varname.startswith('@'):
            raise AppArmorException("Passed invalid varname to %s (doesn't start with '@'): %s" % (self.__class__.__name__, varname))
        if (varname.startswith('@{') and not varname.endswith('}')) or (not varname.startswith('@{') and varname.endswith('}')):
            raise AppArmorException("Passed invalid varname to %s (mismatched braces): %s" % (self.__class__.__name__, varname))

        if not isinstance(mode, str):
            raise AppArmorBug('Passed unknown type for variable assignment mode to %s: %s' % (self.__class__.__name__, mode))
        if mode not in ('=', '+='):
            raise AppArmorBug('Passed unknown variable assignment mode to %s: %s' % (self.__class__.__name__, mode))

        if not isinstance(values, set):
            raise AppArmorBug('Passed unknown type for values to %s: %s' % (self.__class__.__name__, values))
        if not values:
            raise AppArmorException('Passed empty list of values to %s: %s' % (self.__class__.__name__, values))

        self.varname = varname
        self.mode = mode
        self.values = values

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        """parse raw_rule and return instance of this class"""

        comment = parse_comment(matches)

        varname = matches.group('varname')
        mode = matches.group('mode')
        values = separate_vars(matches.group('values'))

        return cls(varname, mode, values,
                   audit=False, deny=False, allow_keyword=False, comment=comment, priority=None)

    def get_clean(self, depth=0):
        """return rule (in clean/default formatting)"""

        space = '  ' * depth

        data = []
        for value in sorted(self.values):
            if not value:
                value = '""'
            data.append(quote_if_needed(value))

        return '%s%s %s %s' % (space, self.varname, self.mode, ' '.join(data))

    def _is_covered_localvars(self, other_rule):
        """check if other_rule is covered by this rule object"""

        if self.varname != other_rule.varname:
            return False

        if self.mode != other_rule.mode:
            return False

        if not self._is_covered_list(self.values, None, set(other_rule.values), None, 'values'):
            return False

        # still here? -> then it is covered
        return True

    def _is_equal_localvars(self, other_rule, strict):
        """compare if rule-specific variables are equal"""

        if self.varname != other_rule.varname:
            return False

        if self.mode != other_rule.mode:
            return False

        if self.values != other_rule.values:
            return False

        return True

    def _logprof_header_localvars(self):
        return _('Variable'), self.get_clean()


class VariableRuleset(BaseRuleset):
    """Class to handle and store a collection of variable rules"""

    def add(self, rule, cleanup=False):
        """Add variable rule object

           If the variable name is already known, raise an exception because re-defining a variable isn't allowed.
        """

        if rule.mode == '=':
            for knownrule in self.rules:
                if rule.varname == knownrule.varname:
                    raise AppArmorException(
                        _('Redefining existing variable %(variable)s: %(value)s')
                        % {'variable': rule.varname, 'value': rule.values})

        super().add(rule, cleanup)

    def get_merged_variables(self):
        """Get merged variables of this object.

           Note that no error checking is done because variables can be defined in one file and extended in another.
        """

        var_set = {}
        var_add = {}

        for rule in self.rules:
            if rule.mode == '=':
                var_set[rule.varname] = rule.values  # blindly set, add() prevents redefinition of variables
            else:
                if not var_add.get(rule.varname):
                    var_add[rule.varname] = rule.values
                else:
                    var_add[rule.varname] |= rule.values

        return {'=': var_set, '+=': var_add}


def separate_vars(vs):
    """Returns a list of all the values for a variable"""
    data = set()
    vs = vs.strip()

    re_vars = re.compile(r'^(("[^"]*")|([^"\s]+))\s*(.*)$')
    while re_vars.search(vs):
        matches = re_vars.search(vs).groups()

        if matches[0].endswith(','):
            raise AppArmorException(_('Variable declarations do not accept trailing commas'))

        data.add(strip_quotes(matches[0]))
        vs = matches[3].strip()

    if vs:
        raise AppArmorException('Variable assignments contains invalid parts (unbalanced quotes?): %s' % vs)

    return data



# ----------------------------------------------------------------------
#    Copyright (C) 2024 Canonical, Ltd.
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------
import re

from apparmor.common import AppArmorBug, AppArmorException

from apparmor.regex import RE_PROFILE_MOUNT, RE_PROFILE_PATH_OR_VAR, strip_parenthesis, strip_quotes, strip_braces
from apparmor.rule import AARE
from apparmor.rule import BaseRule, BaseRuleset, parse_modifiers, logprof_value_or_all, check_and_split_list, quote_if_needed

from apparmor.translations import init_translation

_ = init_translation()

# TODO : Apparmor remount logs are displayed as mount (with remount flag). Profiles generated with aa-genprof are therefore mount rules. It could be interesting to make them remount rules.

flags_bind_mount = {'B', 'bind', 'R', 'rbind'}
flags_change_propagation = {
    'remount', 'unbindable', 'shared', 'private', 'slave', 'runbindable', 'rshared', 'rprivate', 'rslave',
    'make-unbindable', 'make-shared', 'make-private', 'make-slave', 'make-runbindable', 'make-rshared', 'make-rprivate',
    'make-rslave'
}
# keep in sync with parser/mount.cc mnt_opts_table!
flags_keywords = list(flags_bind_mount) + list(flags_change_propagation) + [
    'ro', 'r', 'read-only', 'rw', 'w', 'suid', 'nosuid', 'dev', 'nodev', 'exec', 'noexec', 'sync', 'async', 'mand',
    'nomand', 'dirsync', 'symfollow', 'nosymfollow', 'atime', 'noatime', 'diratime', 'nodiratime', 'move', 'M',
    'verbose', 'silent', 'loud', 'acl', 'noacl', 'relatime', 'norelatime', 'iversion', 'noiversion', 'strictatime',
    'nostrictatime', 'lazytime', 'nolazytime', 'user', 'nouser',
    '[A-Za-z0-9-]+',  # as long as the parser uses a hardcoded options list, this only helps to print a better error message on unknown mount options
]
join_valid_flags = '|'.join(flags_keywords)

sep = r'\s*[\s,]\s*'

# We aim to be a bit more restrictive than \S+ used in regex.py
FS_AARE = r'([][".*@{}\w^-]+)'

fs_type_pattern = r'\b(?P<fstype_or_vfstype>fstype|vfstype)\b\s*(?P<fstype_equals_or_in>=|in)\s*'\
    r'(?P<fstype>\(\s*(' + FS_AARE + r')(' + sep + r'(' + FS_AARE + r'))*\s*\)|'\
    r'\{\s*(' + FS_AARE + r')(' + sep + r'(' + FS_AARE + r'))*\s*\}|'\
    r'(\s*' + FS_AARE + r')(,' + FS_AARE + ')*)'


option_pattern = r'\s*(\boption(s?)\b\s*(?P<options_equals_or_in>=|in)\s*'\
    r'(?P<options>\(\s*(' + join_valid_flags + r')(' + sep + r'(' + join_valid_flags + r'))*\s*\)|' \
    r'(\s*' + join_valid_flags + r')(?=(\s|$))'\
    r'))'

# allow any order of fstype and options
# Note: also matches if multiple fstype= or options= are given to keep the regex simpler
mount_condition_pattern = rf'({fs_type_pattern}\s*|{option_pattern}\s*)*'

# Source can either be
# - A path          : /foo
# - A globbed Path  : {,/usr}/lib{,32,64,x32}/modules/
# - A filesystem    : sysfs         (sudo mount -t tmpfs tmpfs /tmp/bar)
# - Any label       : mntlabel      (sudo mount -t tmpfs mntlabel /tmp/bar)
# Thus we cannot use directly RE_PROFILE_PATH_OR_VAR
# Destination can also be
# - A path          : /foo
# - A globbed Path  : **

glob_pattern = (
    r'(\s*(?P<%s>('
    + RE_PROFILE_PATH_OR_VAR % 'IGNOREDEV'  # path or variable
    + r'|\{\S*|"\{[^"]*"'  # alternation, optionally quoted (note: no leading "/" needed/enforced)
    + r'|\*\*\S*|\*\*[^"]*"'  # starting with "**"
    # Note: the closing ')))' needs to be added in the final regex
)

source_fileglob_pattern = (
    glob_pattern % 'source_file'
    + r'|""'  # empty source
    + r'|[\w-]+'  # any word including "-"
    + ')))'
)

dest_fileglob_pattern = (
    glob_pattern.replace('IGNOREDEV', 'IGNOREMP') % 'dest_file'
    + ')))'
)

RE_MOUNT_DETAILS = re.compile(r'^\s*' + mount_condition_pattern + rf'(\s+{source_fileglob_pattern})?' + rf'(\s+->\s+{dest_fileglob_pattern})?\s*' + r'$')
RE_UMOUNT_DETAILS = re.compile(r'^\s*' + mount_condition_pattern + rf'(\s+{dest_fileglob_pattern})?\s*' + r'$')

# check if a rule contains multiple 'fstype'
# (not using fs_type_pattern here because a) it also matches an empty string, and b) using it twice would cause name conflicts)
multi_param_template = r'\sPARAM\s*(=|\sin).*\sPARAM\s*(=|\sin)'
RE_MOUNT_MULTIPLE_FS_TYPE = re.compile(multi_param_template.replace('PARAM', 'v?fstype'))


class MountRule(BaseRule):
    '''Class to handle and store a single mount rule'''

    # Nothing external should reference this class, all external users
    # should reference the class field MountRule.ALL
    class __MountAll(object):
        pass

    ALL = __MountAll

    rule_name = 'mount'
    _match_re = RE_PROFILE_MOUNT

    def __init__(self, operation, fstype, options, source, dest,
                 audit=False, deny=False, allow_keyword=False,
                 comment='', log_event=None, priority=None):

        super().__init__(audit=audit, deny=deny,
                         allow_keyword=allow_keyword,
                         comment=comment,
                         log_event=log_event,
                         priority=priority)

        self.operation = operation

        if not isinstance(fstype, list):
            fstype = [fstype]

        # self.all_fstype will only be true if no fstypes are
        # specified, so it's fine to set it inside the loop
        self.fstype = []
        for fst in fstype:
            if fst == self.ALL or fst[1] == self.ALL:
                self.all_fstype = True
                fstype_values = None
                is_fstype_equal = None
            else:
                self.all_fstype = False
                for it in fst[1]:
                    aare_len, unused = parse_aare(it, 0, 'fstype')
                    if aare_len != len(it):
                        raise AppArmorException(f'Invalid aare : {it}')
                fstype_values = fst[1]
                is_fstype_equal = fst[0]
            self.fstype.append(MountConditional('fstype', fstype_values, self.all_fstype, is_fstype_equal, 'aare'))

        if not isinstance(options, list):
            options = [options]

        # self.all_options will only be true if no options are
        # specified, so it's fine to set it inside the loop
        self.options = []
        for opts in options:
            opt_values, self.all_options, unknown_items = check_and_split_list(opts[1] if opts != self.ALL else opts, flags_keywords, self.ALL, type(self).__name__, 'options')
            if unknown_items:
                raise AppArmorException(_('Passed unknown options keyword to %s: %s') % (type(self).__name__, ' '.join(unknown_items)))
            is_options_equal = opts[0] if not self.all_options else None
            self.options.append(MountConditional('options', opt_values, self.all_options, is_options_equal, 'list'))

        self.source, self.all_source = self._aare_or_all(source, 'source', is_path=False, log_event=log_event, empty_ok=True)
        self.dest, self.all_dest = self._aare_or_all(dest, 'dest', is_path=False, log_event=log_event)

        if self.operation != 'mount' and not self.all_source:
            raise AppArmorException(f'Operation {self.operation} cannot have a source')

        for opt_cond in self.options:
            if self.operation == 'mount' and not self.all_options and flags_change_propagation & opt_cond.values != set():
                if not (self.all_source or self.all_dest):
                    raise AppArmorException(f'Operation {flags_change_propagation & opt_cond.values} cannot specify a source. Source = {self.source}')
                elif not self.all_fstype:
                    raise AppArmorException(f'Operation {flags_change_propagation & opt_cond.values} cannot specify a fstype. Fstype = {self.fstype}')

            if self.operation == 'mount' and not self.all_options and flags_bind_mount & opt_cond.values != set() and not self.all_fstype:
                raise AppArmorException(f'Bind mount rules cannot specify a fstype. Fstype = {self.fstype}')

        self.can_glob = not self.all_source and not self.all_dest and not self.all_options

    @classmethod
    def _create_instance(cls, raw_rule, matches):
        '''parse raw_rule and return instance of this class'''

        priority, audit, deny, allow_keyword, comment = parse_modifiers(matches)

        operation = matches.group('operation')

        rule_details = ''
        if matches.group('details'):
            rule_details = matches.group('details')

            if operation == 'mount':
                parsed = RE_MOUNT_DETAILS.search(rule_details)
            else:
                parsed = RE_UMOUNT_DETAILS.search(rule_details)

            r = parsed.groupdict() if parsed else None
            if not r:
                raise AppArmorException('Can\'t parse mount rule ' + raw_rule)

            fstype = (None, cls.ALL)
            if r['fstype'] is not None:
                fstype = []
                for m in re.finditer(fs_type_pattern, rule_details):
                    # apparmor_parser cannot handle items starting with '{' so '({ext3,ext4})' will fail -> Converting to (ext3,ext4)
                    fst = parse_aare_list(strip_braces(strip_parenthesis(m.group('fstype')).strip()), 'fstype')
                    fstype.append((m.group('fstype_equals_or_in'), fst))

            opts = (None, cls.ALL)
            if r['options'] is not None:
                opts = []
                for m in re.finditer(option_pattern, rule_details):
                    options = strip_parenthesis(m.group('options')).replace(',', ' ').split()
                    opts.append((m.group('options_equals_or_in'), options))

            if operation == 'mount' and r['source_file'] is not None:  # Umount cannot have a source
                source = strip_quotes(r['source_file'])
            else:
                source = cls.ALL

            if r['dest_file'] is not None:
                dest = strip_quotes(r['dest_file'])
            else:
                dest = cls.ALL

        else:
            opts = (None, cls.ALL)
            fstype = (None, cls.ALL)
            source = cls.ALL
            dest = cls.ALL

        return cls(operation=operation, fstype=fstype, options=opts,
                   source=source, dest=dest, audit=audit, deny=deny, allow_keyword=allow_keyword, comment=comment,
                   priority=priority)

    def get_clean(self, depth=0):
        space = '  ' * depth

        fstype = ''
        for fst in self.fstype:
            fstype += fst.get_clean()

        options = ''
        for opt in self.options:
            options += opt.get_clean()

        source = ''
        dest = ''

        if self.operation == 'mount':
            if not self.all_source:
                if self.source.regex == '':
                    source = ' ""'
                else:
                    source = ' ' + quote_if_needed(str(self.source.regex))

            if not self.all_dest:
                dest = ' -> ' + quote_if_needed(str(self.dest.regex))

        else:
            if not self.all_dest:
                dest = ' ' + quote_if_needed(str(self.dest.regex))

        return ('%s%s%s%s%s%s%s,%s' % (space,
                                       self.modifiers_str(),
                                       self.operation,
                                       fstype,
                                       options,
                                       source,
                                       dest,
                                       self.comment,
                                       ))

    def _is_cond_list_covered(self, conds, other_conds):
        '''Checks if all conds in 'other_conds' are covered by at
        least one cond in 'conds'.'''
        return all(
            any(cond.is_covered(other_cond) for cond in conds)
            for other_cond in other_conds
        )

    def _is_covered_localvars(self, other_rule):
        if self.operation != other_rule.operation:
            return False
        if not self._is_cond_list_covered(self.fstype, other_rule.fstype):
            return False
        if not self._is_cond_list_covered(self.options, other_rule.options):
            return False
        if not self._is_covered_aare(self.source, self.all_source, other_rule.source, other_rule.all_source, 'source'):
            return False
        if not self._is_covered_aare(self.dest, self.all_dest, other_rule.dest, other_rule.all_dest, 'dest'):
            return False

        return True

    def _is_equal_localvars(self, other_rule, strict):
        if self.operation != other_rule.operation:
            return False
        if self.fstype != other_rule.fstype or self.options != other_rule.options:
            return False
        if not self._is_equal_aare(self.source, self.all_source, other_rule.source, other_rule.all_source, 'source'):
            return False
        if not self._is_equal_aare(self.dest, self.all_dest, other_rule.dest, other_rule.all_dest, 'dest'):
            return False

        return True

    @staticmethod
    def hashlog_from_event(hl, ev):
        if ev['flags'] is not None:
            ev['flags'] = ('=', ev['flags'])
        if ev['fs_type'] is not None:
            ev['fs_type'] = ('=', ev['fs_type'])
        if ev['operation'] == 'mount':
            hl[ev['operation']][ev['flags']][ev['fs_type']][ev['name']][ev['src_name']] = True
        else:  # Umount
            hl[ev['operation']][ev['flags']][ev['fs_type']][ev['name']][None] = True

    @classmethod
    def from_hashlog(cls, hl):
        for operation, options, fstype, dest, source in cls.generate_rules_from_hashlog(hl, 5):
            _options = (options[0], options[1].split(', ')) if options is not None else MountRule.ALL
            _fstype = (fstype[0], fstype[1].split(', ')) if fstype is not None else MountRule.ALL
            _source = source if source is not None else MountRule.ALL
            _dest = dest if dest is not None else MountRule.ALL
            yield cls(operation=operation, fstype=_fstype, options=_options, source=_source, dest=_dest)

    def glob(self):
        '''Change path to next possible glob'''
        if self.all_source and self.all_options:
            return

        if not self.all_dest:
            self.all_dest = True
            self.dest = self.ALL
        elif not self.all_source and type(self.source) is not str:
            self.source = self.source.glob_path()
            if self.source.is_equal('/**/'):
                self.all_source = True
                self.source = self.ALL

        else:
            self.options = [MountConditional('options', self.ALL, True, None)]
            self.all_options = True
        self.raw_rule = None

    def _logprof_header_localvars(self):
        operation = self.operation

        fstype_output = ()
        for fst in self.fstype:
            fstype = logprof_value_or_all(fst.values, fst.all_values)
            fstype_output = fstype_output + (_('Fstype'), (fst.operator, fstype) if fstype != 'ALL' else fstype)

        opts_output = ()
        for opt in self.options:
            options = logprof_value_or_all(opt.values, opt.all_values)
            opts_output = opts_output + (_('Options'), (opt.operator, options) if options != 'ALL' else options)
        source = logprof_value_or_all(self.source, self.all_source)
        dest = logprof_value_or_all(self.dest, self.all_dest)

        return (
            _('Operation'), operation,
            *fstype_output,
            *opts_output,
            _('Source'), source,
            _('Destination'), dest,

        )


class MountRuleset(BaseRuleset):
    '''Class to handle and store a collection of Mount rules'''


def parse_aare(s, offset, param):
    parsed = ''
    brace_count = 0
    for i, c in enumerate(s[offset:], start=offset):
        if c in [' ', ',', '\t'] and brace_count == 0:
            break
        parsed += c
        if c == '{':
            brace_count += 1
        elif c == '}':
            brace_count -= 1
            if brace_count < 0:
                raise AppArmorException(f"Unmatched closing brace in {param}: {s[offset:]}")
        offset = i

    if brace_count != 0:
        raise AppArmorException(f"Unmatched opening brace in {param}: {s[offset:]}")

    return offset + 1, parsed


def parse_aare_list(s, param):
    res = []
    offset = 0
    while offset <= len(s):
        offset, part = parse_aare(s, offset, param)
        if part.translate(' ,\t') != '':
            res.append(part)
    return res


def wrap_in_with_spaces(value):
    ''' wrap 'in' keyword in spaces, and leave everything else unchanged '''

    if value == 'in':
        value = ' in '

    return value


class MountConditional(MountRule):
    '''Class to handle and store mount conditionals'''
    def __init__(self, name, values, all_values, operator, cond_type=None):
        self.name = name
        self.values = values
        self.all_values = all_values
        self.operator = operator
        self.cond_type = cond_type
        self.raw_rule = ''  # needed so __repr__ calls get_clean

        if not self.all_values and self.operator not in ('=', 'in'):
            raise AppArmorBug(f'Invalid operator for {self.name}: {self.operator}')

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MountConditional):
            return False

        if self.all_values != other.all_values:
            return False

        if self.cond_type != other.cond_type:
            return False

        if self.values != other.values:
            return False

        if self.operator != other.operator:
            return False

        return True

    def is_covered(self, other: object) -> bool:
        if not isinstance(other, MountConditional):
            return False

        if self.name != other.name:
            return False

        if self.all_values:
            return True

        if other.all_values:
            return False

        if self.cond_type == 'list':
            if not self._is_covered_list(self.values, self.all_values, other.values, other.all_values, self.name):
                return False
        elif self.cond_type == 'aare':  # list of aares - all values in other must be at least once in self.values
            if not all(any(self._is_covered_aare(AARE(value, False), self.all_values,
                                                 AARE(other_value, False), other.all_values, self.name)
                           for value in self.values)
                       for other_value in other.values):
                return False
        else:
            raise AppArmorBug('Type should only be empty if ALL is true')

        if self.operator == other.operator:
            return True

        return False

    def get_clean(self, depth=0) -> str:
        conditional = ''
        if not self.all_values:
            if self.name == 'fstype' and self.operator == "=":
                conditional += ' %s%s%s' % (self.name, wrap_in_with_spaces(self.operator), ','.join(sorted(self.values)))
            else:
                conditional += ' %s%s(%s)' % (self.name, wrap_in_with_spaces(self.operator), ', '.join(sorted(self.values)))

        return conditional



# ----------------------------------------------------------------------
#    Copyright (C) 2018-2024 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

from apparmor.aare import AARE
from apparmor.common import AppArmorBug, AppArmorException
from apparmor.profile_storage import ProfileStorage
from apparmor.rule.abi import AbiRule, AbiRuleset
from apparmor.rule.alias import AliasRule, AliasRuleset
from apparmor.rule.boolean import BooleanRule, BooleanRuleset
from apparmor.rule.include import IncludeRule, IncludeRuleset
from apparmor.rule.variable import VariableRule, VariableRuleset
from apparmor.translations import init_translation

_ = init_translation()

preamble_ruletypes = {
    'abi':      {'rule': AbiRule,      'ruleset': AbiRuleset},
    'alias':    {'rule': AliasRule,    'ruleset': AliasRuleset},
    'inc_ie':   {'rule': IncludeRule,  'ruleset': IncludeRuleset},
    'variable': {'rule': VariableRule, 'ruleset': VariableRuleset},
    'boolean':  {'rule': BooleanRule,  'ruleset': BooleanRuleset},
}
header_rule_write_order = ('abi', 'alias', 'inc_ie', 'variable', 'boolean')  # TODO: Dicts are ordered in Python 3.7+; use above dict's keys instead


class ProfileList:
    """Stores the preamble section and the list of profile(s) (both name and
       attachment) that live in profile files.

       Also allows "reverse" lookups to find out in which file a profile
       lives.
    """

    def __init__(self):
        self.profile_names = {}     # profile name -> filename
        self.attachments = {}       # attachment -> {'f': filename, 'p': profile, 're': AARE(attachment)}
        self.files = {}             # filename -> content - see init_file()
        self.profiles = {}          # profile_name -> ProfileStorage

    def __repr__(self):
        name = type(self).__name__
        return '\n<%s>\n%s\n</%s>\n' % (name, '\n'.join(self.files), name)

    def __getitem__(self, key):
        if key in self.profiles:
            return self.profiles[key]
        else:
            raise AppArmorBug('attempt to read unknown profile %s' % key)

    def init_file(self, filename):
        if self.files.get(filename):
            return  # don't re-initialize / overwrite existing data

        self.files[filename] = {
            'profiles': [],
        }

        for rule in preamble_ruletypes:
            self.files[filename][rule] = preamble_ruletypes[rule]['ruleset']()

    def add_profile(self, filename, profile_name, attachment, prof_storage):
        """Add the given profile and attachment to the list"""

        if not filename:
            raise AppArmorBug('Empty filename given to ' + type(self).__name__)

        if not profile_name and not attachment:
            raise AppArmorBug('Neither profile name or attachment given')

        if type(prof_storage) is not ProfileStorage:
            raise AppArmorBug('Invalid profile type: %s' % type(prof_storage))

        if profile_name in self.profile_names:
            raise AppArmorException(
                _('Profile %(profile_name)s exists in %(filename)s and %(filename2)s'
                  % {'profile_name': profile_name, 'filename': filename, 'filename2': self.profile_names[profile_name]}))

        if attachment in self.attachments:
            raise AppArmorException(
                _('Profile for %(profile_name)s exists in %(filename)s and %(filename2)s'
                  % {'profile_name': attachment, 'filename': filename, 'filename2': self.attachments[attachment]}))

        if profile_name:
            self.profile_names[profile_name] = filename

        if attachment:
            self.attachments[attachment] = {'f': filename, 'p': profile_name or attachment}  # if a profile doesn't have a name, the attachment is stored as profile name
            self.attachments[attachment]['re'] = AARE(attachment, True)

        self.init_file(filename)

        if profile_name:
            self.files[filename]['profiles'].append(profile_name)
            self.profiles[profile_name] = prof_storage
        else:
            self.files[filename]['profiles'].append(attachment)
            self.profiles[attachment] = prof_storage

    def replace_profile(self, profile_name, prof_storage):
        """Replace the given profile in the profile list"""

        if profile_name not in self.profiles:
            raise AppArmorBug('Attempt to replace non-existing profile %s' % profile_name)

        if type(prof_storage) is not ProfileStorage:
            raise AppArmorBug('Invalid profile type: %s' % type(prof_storage))

        # we might lift this restriction later, but for now, forbid changing the attachment instead of updating self.attachments
        if self.profiles[profile_name]['attachment'] != prof_storage['attachment']:
            raise AppArmorBug('Attempt to change atttachment while replacing profile %s - original: %s, new: %s' % (profile_name, self.profiles[profile_name]['attachment'], prof_storage['attachment']))

        self.profiles[profile_name] = prof_storage

    def add_rule(self, filename, ruletype, rule):
        """Store the given rule for the given profile filename preamble"""

        self.init_file(filename)

        self.files[filename][ruletype].add(rule)

    def add_abi(self, filename, abi_rule):
        """Store the given abi rule for the given profile filename preamble"""

        if type(abi_rule) is not AbiRule:
            raise AppArmorBug('Wrong type given to %s: %s' % (type(self).__name__, abi_rule))

        self.init_file(filename)

        self.files[filename]['abi'].add(abi_rule)

    def add_alias(self, filename, alias_rule):
        """Store the given alias rule for the given profile filename preamble"""

        if type(alias_rule) is not AliasRule:
            raise AppArmorBug('Wrong type given to %s: %s' % (type(self).__name__, alias_rule))

        self.init_file(filename)

        self.files[filename]['alias'].add(alias_rule)

    def add_inc_ie(self, filename, inc_rule):
        """Store the given include / include if exists rule for the given profile filename preamble"""
        if type(inc_rule) is not IncludeRule:
            raise AppArmorBug('Wrong type given to %s: %s' % (type(self).__name__, inc_rule))

        self.init_file(filename)

        self.files[filename]['inc_ie'].add(inc_rule)

    def add_variable(self, filename, var_rule):
        """Store the given variable rule for the given profile filename preamble"""
        if type(var_rule) is not VariableRule:
            raise AppArmorBug('Wrong type given to %s: %s' % (type(self).__name__, var_rule))

        self.init_file(filename)

        self.files[filename]['variable'].add(var_rule)

    def add_boolean(self, filename, bool_rule):
        """Store the given boolean variable rule for the given profile filename preamble"""
        if type(bool_rule) is not BooleanRule:
            raise AppArmorBug('Wrong type given to %s: %s' % (type(self).__name__, bool_rule))

        self.init_file(filename)

        self.files[filename]['boolean'].add(bool_rule)

    def delete_preamble_duplicates(self, filename):
        """Delete duplicates in the preamble of the given profile file"""

        if not self.files.get(filename):
            raise AppArmorBug('%s not listed in %s files' % (filename, type(self).__name__))

        deleted = 0

        for r_type in preamble_ruletypes:
            deleted += self.files[filename][r_type].delete_duplicates(None)  # None means not to check includes -- TODO check if this makes sense for all preamble rule types

        return deleted

    def get_all_profiles(self):
        return self.profiles

    def get_profile_and_childs(self, profile_name):
        found = {}
        for prof in self.profiles:
            if prof == profile_name or prof.startswith('%s//' % profile_name):
                found[prof] = self.profiles[prof]

        return found

    def get_raw(self, filename, depth=0):
        """Get the preamble for the given profile filename (in original formatting)"""
        if not self.files.get(filename):
            raise AppArmorBug('%s not listed in %s files' % (filename, type(self).__name__))

        data = []
        for rule_type in header_rule_write_order:
            data.extend(self.files[filename][rule_type].get_raw(depth))
        return data

    def get_clean(self, filename, depth=0):
        """Get the preamble for the given profile filename (in clean formatting)"""
        if not self.files.get(filename):
            raise AppArmorBug('%s not listed in %s files' % (filename, type(self).__name__))

        data = []
        for rule_type in header_rule_write_order:
            data.extend(self.files[filename][rule_type].get_clean_unsorted(depth))
        return data

    def filename_from_profile_name(self, name):
        """Return profile filename for the given profile name, or None"""

        return self.profile_names.get(name, None)

    def filename_from_attachment(self, attachment):
        """Return profile filename for the given attachment/executable path, or None"""
        return self.thing_from_attachment(attachment, 'f')

    def profile_from_attachment(self, attachment):
        """Return profile filename for the given attachment/executable path, or None"""
        return self.thing_from_attachment(attachment, 'p')

    def thing_from_attachment(self, attachment, thing):
        """Return thing for the given attachment/executable path, or None.

           thing can be 'f' for filename or 'p' for profile name"""

        if not attachment.startswith(('/', '@', '{')):
            raise AppArmorBug('Called filename_from_attachment with non-path attachment: %s' % attachment)

        # plain path
        if self.attachments.get(attachment):
            return self.attachments[attachment][thing]

        # try AARE matches to cover profile names with alternations and wildcards
        for path in self.attachments.keys():
            if self.attachments[path]['re'].match(attachment):
                return self.attachments[path][thing]  # XXX this returns the first match, not necessarily the best one

        return None  # nothing found

    def get_all_merged_variables(self, filename, all_incfiles):
        """Get merged variables of a file and its includes

           Note that this function is more forgiving than apparmor_parser.
           It detects variable redefinitions and adding values to non-existing variables.
           However, it doesn't honor the order - so adding to a variable first and defining
           it later won't trigger an error.
        """

        if not self.files.get(filename):
            raise AppArmorBug('%s not listed in %s files' % (filename, type(self).__name__))

        merged_variables = {}

        mainfile_variables = self.files[filename]['variable'].get_merged_variables()

        # keep track in which file a variable gets set
        set_in = {}
        for var in mainfile_variables['=']:
            merged_variables[var] = mainfile_variables['='][var]
            set_in[var] = filename

        # collect variable additions (+=)
        inc_add = {}
        if mainfile_variables['+=']:
            inc_add[filename] = mainfile_variables['+=']  # variable additions from main file

        for incname in all_incfiles:
            inc_vars = self.files[incname]['variable'].get_merged_variables()

            for var in inc_vars['=']:
                if merged_variables.get(var):
                    raise AppArmorException(
                        'While parsing %(profile)s: Conflicting variable definitions for variable %(var)s found in %(file1)s and %(file2)s.'
                        % {'var': var, 'profile': filename, 'file1': set_in[var], 'file2': incname})
                else:
                    merged_variables[var] = inc_vars['='][var]
                    set_in[var] = incname

            # variable additions can happen in other files than the variable definition. First collect them from all files...
            if inc_vars['+=']:
                inc_add[incname] = inc_vars['+=']

        for incname in inc_add:
            # ... and then check if the variables that get extended have an initial definition. If yes, merge them.
            for var in inc_add[incname]:
                if merged_variables.get(var):
                    merged_variables[var] |= inc_add[incname][var]
                else:
                    raise AppArmorException(
                        'While parsing %(profile)s: Variable %(var)s was not previously declared, but is being assigned additional value in file %(file)s.'
                        % {'var': var, 'profile': filename, 'file': incname})

        return merged_variables

    def profile_exists(self, profile_name):
        return profile_name in self.profiles

    def profiles_in_file(self, filename):
        """Return list of profiles in the given file"""
        if not self.files.get(filename):
            raise AppArmorBug('%s not listed in %s files' % (filename, type(self).__name__))

        return self.files[filename]['profiles']



# ------------------------------------------------------------------
#
#    Copyright (C) 2011-2012 Canonical Ltd.
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License published by the Free Software Foundation.
#
# ------------------------------------------------------------------



# This file was automatically generated by SWIG (https://www.swig.org).
# Version 4.4.1
#
# Do not make changes to this file unless you know what you are doing - modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
# Import the low-level C/C++ module
if getattr(globals().get("__spec__"), "parent", None) or __package__ or "." in __name__:
    from . import _LibAppArmor
else:
    import _LibAppArmor

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "this":
            set(self, name, value)
        elif name == "thisown":
            self.this.own(value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)



def free_record(record):
    r"""This function used to free aa_log_record objects. Freeing is now handled automatically, so this no-op function remains for backwards compatibility."""

    import warnings
    warnings.warn("free_record is now a no-op as the record's memory is handled automatically", DeprecationWarning)


    return _LibAppArmor.free_record(record)
AA_RECORD_EXEC_MMAP = _LibAppArmor.AA_RECORD_EXEC_MMAP
AA_RECORD_READ = _LibAppArmor.AA_RECORD_READ
AA_RECORD_WRITE = _LibAppArmor.AA_RECORD_WRITE
AA_RECORD_EXEC = _LibAppArmor.AA_RECORD_EXEC
AA_RECORD_LINK = _LibAppArmor.AA_RECORD_LINK
AA_RECORD_SYNTAX_V1 = _LibAppArmor.AA_RECORD_SYNTAX_V1
AA_RECORD_SYNTAX_V2 = _LibAppArmor.AA_RECORD_SYNTAX_V2
AA_RECORD_SYNTAX_UNKNOWN = _LibAppArmor.AA_RECORD_SYNTAX_UNKNOWN
AA_RECORD_INVALID = _LibAppArmor.AA_RECORD_INVALID
AA_RECORD_ERROR = _LibAppArmor.AA_RECORD_ERROR
AA_RECORD_AUDIT = _LibAppArmor.AA_RECORD_AUDIT
AA_RECORD_ALLOWED = _LibAppArmor.AA_RECORD_ALLOWED
AA_RECORD_DENIED = _LibAppArmor.AA_RECORD_DENIED
AA_RECORD_HINT = _LibAppArmor.AA_RECORD_HINT
AA_RECORD_STATUS = _LibAppArmor.AA_RECORD_STATUS
class aa_log_record(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    version = property(_LibAppArmor.aa_log_record_version_get, _LibAppArmor.aa_log_record_version_set)
    event = property(_LibAppArmor.aa_log_record_event_get, _LibAppArmor.aa_log_record_event_set)
    pid = property(_LibAppArmor.aa_log_record_pid_get, _LibAppArmor.aa_log_record_pid_set)
    peer_pid = property(_LibAppArmor.aa_log_record_peer_pid_get, _LibAppArmor.aa_log_record_peer_pid_set)
    task = property(_LibAppArmor.aa_log_record_task_get, _LibAppArmor.aa_log_record_task_set)
    magic_token = property(_LibAppArmor.aa_log_record_magic_token_get, _LibAppArmor.aa_log_record_magic_token_set)
    epoch = property(_LibAppArmor.aa_log_record_epoch_get, _LibAppArmor.aa_log_record_epoch_set)
    audit_sub_id = property(_LibAppArmor.aa_log_record_audit_sub_id_get, _LibAppArmor.aa_log_record_audit_sub_id_set)
    bitmask = property(_LibAppArmor.aa_log_record_bitmask_get, _LibAppArmor.aa_log_record_bitmask_set)
    audit_id = property(_LibAppArmor.aa_log_record_audit_id_get, _LibAppArmor.aa_log_record_audit_id_set)
    operation = property(_LibAppArmor.aa_log_record_operation_get, _LibAppArmor.aa_log_record_operation_set)
    denied_mask = property(_LibAppArmor.aa_log_record_denied_mask_get, _LibAppArmor.aa_log_record_denied_mask_set)
    requested_mask = property(_LibAppArmor.aa_log_record_requested_mask_get, _LibAppArmor.aa_log_record_requested_mask_set)
    fsuid = property(_LibAppArmor.aa_log_record_fsuid_get, _LibAppArmor.aa_log_record_fsuid_set)
    ouid = property(_LibAppArmor.aa_log_record_ouid_get, _LibAppArmor.aa_log_record_ouid_set)
    profile = property(_LibAppArmor.aa_log_record_profile_get, _LibAppArmor.aa_log_record_profile_set)
    peer_profile = property(_LibAppArmor.aa_log_record_peer_profile_get, _LibAppArmor.aa_log_record_peer_profile_set)
    comm = property(_LibAppArmor.aa_log_record_comm_get, _LibAppArmor.aa_log_record_comm_set)
    name = property(_LibAppArmor.aa_log_record_name_get, _LibAppArmor.aa_log_record_name_set)
    name2 = property(_LibAppArmor.aa_log_record_name2_get, _LibAppArmor.aa_log_record_name2_set)
    namespace = property(_LibAppArmor.aa_log_record_namespace_get, _LibAppArmor.aa_log_record_namespace_set)
    attribute = property(_LibAppArmor.aa_log_record_attribute_get, _LibAppArmor.aa_log_record_attribute_set)
    parent = property(_LibAppArmor.aa_log_record_parent_get, _LibAppArmor.aa_log_record_parent_set)
    info = property(_LibAppArmor.aa_log_record_info_get, _LibAppArmor.aa_log_record_info_set)
    peer_info = property(_LibAppArmor.aa_log_record_peer_info_get, _LibAppArmor.aa_log_record_peer_info_set)
    error_code = property(_LibAppArmor.aa_log_record_error_code_get, _LibAppArmor.aa_log_record_error_code_set)
    active_hat = property(_LibAppArmor.aa_log_record_active_hat_get, _LibAppArmor.aa_log_record_active_hat_set)
    net_family = property(_LibAppArmor.aa_log_record_net_family_get, _LibAppArmor.aa_log_record_net_family_set)
    net_protocol = property(_LibAppArmor.aa_log_record_net_protocol_get, _LibAppArmor.aa_log_record_net_protocol_set)
    net_sock_type = property(_LibAppArmor.aa_log_record_net_sock_type_get, _LibAppArmor.aa_log_record_net_sock_type_set)
    net_local_addr = property(_LibAppArmor.aa_log_record_net_local_addr_get, _LibAppArmor.aa_log_record_net_local_addr_set)
    net_local_port = property(_LibAppArmor.aa_log_record_net_local_port_get, _LibAppArmor.aa_log_record_net_local_port_set)
    net_foreign_addr = property(_LibAppArmor.aa_log_record_net_foreign_addr_get, _LibAppArmor.aa_log_record_net_foreign_addr_set)
    net_foreign_port = property(_LibAppArmor.aa_log_record_net_foreign_port_get, _LibAppArmor.aa_log_record_net_foreign_port_set)
    dbus_bus = property(_LibAppArmor.aa_log_record_dbus_bus_get, _LibAppArmor.aa_log_record_dbus_bus_set)
    dbus_path = property(_LibAppArmor.aa_log_record_dbus_path_get, _LibAppArmor.aa_log_record_dbus_path_set)
    dbus_interface = property(_LibAppArmor.aa_log_record_dbus_interface_get, _LibAppArmor.aa_log_record_dbus_interface_set)
    dbus_member = property(_LibAppArmor.aa_log_record_dbus_member_get, _LibAppArmor.aa_log_record_dbus_member_set)
    signal = property(_LibAppArmor.aa_log_record_signal_get, _LibAppArmor.aa_log_record_signal_set)
    peer = property(_LibAppArmor.aa_log_record_peer_get, _LibAppArmor.aa_log_record_peer_set)
    fs_type = property(_LibAppArmor.aa_log_record_fs_type_get, _LibAppArmor.aa_log_record_fs_type_set)
    flags = property(_LibAppArmor.aa_log_record_flags_get, _LibAppArmor.aa_log_record_flags_set)
    src_name = property(_LibAppArmor.aa_log_record_src_name_get, _LibAppArmor.aa_log_record_src_name_set)
    _class = property(_LibAppArmor.aa_log_record__class_get, _LibAppArmor.aa_log_record__class_set)
    net_addr = property(_LibAppArmor.aa_log_record_net_addr_get, _LibAppArmor.aa_log_record_net_addr_set)
    peer_addr = property(_LibAppArmor.aa_log_record_peer_addr_get, _LibAppArmor.aa_log_record_peer_addr_set)
    execpath = property(_LibAppArmor.aa_log_record_execpath_get, _LibAppArmor.aa_log_record_execpath_set)
    __swig_destroy__ = _LibAppArmor.delete_aa_log_record

    def __init__(self):
        _LibAppArmor.aa_log_record_swiginit(self, _LibAppArmor.new_aa_log_record())

# Register aa_log_record in _LibAppArmor:
_LibAppArmor.aa_log_record_swigregister(aa_log_record)

def parse_record(str):
    return _LibAppArmor.parse_record(str)

def aa_get_lsm_iface_name(op):
    return _LibAppArmor.aa_get_lsm_iface_name(op)

def aa_splitcon(con):
    return _LibAppArmor.aa_splitcon(con)

def aa_is_enabled():
    return _LibAppArmor.aa_is_enabled()

def aa_change_hat(subprofile, magic_token):
    return _LibAppArmor.aa_change_hat(subprofile, magic_token)

def aa_change_profile(profile):
    return _LibAppArmor.aa_change_profile(profile)

def aa_change_onexec(profile):
    return _LibAppArmor.aa_change_onexec(profile)

def aa_change_hatv(subprofiles, token):
    return _LibAppArmor.aa_change_hatv(subprofiles, token)

def aa_stack_profile(profile):
    return _LibAppArmor.aa_stack_profile(profile)

def aa_stack_onexec(profile):
    return _LibAppArmor.aa_stack_onexec(profile)

def aa_find_mountpoint():
    return _LibAppArmor.aa_find_mountpoint()

def aa_getprocattr(tid, attr):
    return _LibAppArmor.aa_getprocattr(tid, attr)

def aa_get_self_attr(op_type):
    return _LibAppArmor.aa_get_self_attr(op_type)

def aa_set_self_attr(op_type, buf, len):
    return _LibAppArmor.aa_set_self_attr(op_type, buf, len)

def aa_gettaskcon(target):
    return _LibAppArmor.aa_gettaskcon(target)

def aa_getcon():
    return _LibAppArmor.aa_getcon()

def aa_getpeercon(fd):
    return _LibAppArmor.aa_getpeercon(fd)
AA_MAY_EXEC = _LibAppArmor.AA_MAY_EXEC
AA_MAY_WRITE = _LibAppArmor.AA_MAY_WRITE
AA_MAY_READ = _LibAppArmor.AA_MAY_READ
AA_MAY_APPEND = _LibAppArmor.AA_MAY_APPEND
AA_MAY_CREATE = _LibAppArmor.AA_MAY_CREATE
AA_MAY_DELETE = _LibAppArmor.AA_MAY_DELETE
AA_MAY_OPEN = _LibAppArmor.AA_MAY_OPEN
AA_MAY_RENAME = _LibAppArmor.AA_MAY_RENAME
AA_MAY_SETATTR = _LibAppArmor.AA_MAY_SETATTR
AA_MAY_GETATTR = _LibAppArmor.AA_MAY_GETATTR
AA_MAY_SETCRED = _LibAppArmor.AA_MAY_SETCRED
AA_MAY_GETCRED = _LibAppArmor.AA_MAY_GETCRED
AA_MAY_CHMOD = _LibAppArmor.AA_MAY_CHMOD
AA_MAY_CHOWN = _LibAppArmor.AA_MAY_CHOWN
AA_MAY_LOCK = _LibAppArmor.AA_MAY_LOCK
AA_EXEC_MMAP = _LibAppArmor.AA_EXEC_MMAP
AA_MAY_LINK = _LibAppArmor.AA_MAY_LINK
AA_MAY_ONEXEC = _LibAppArmor.AA_MAY_ONEXEC
AA_MAY_CHANGE_PROFILE = _LibAppArmor.AA_MAY_CHANGE_PROFILE

def aa_query_file_path(mask, label, path):
    return _LibAppArmor.aa_query_file_path(mask, label, path)

def aa_query_link_path(label, target, link):
    return _LibAppArmor.aa_query_link_path(label, target, link)




# ----------------------------------------------------------------------
#    Copyright (C) 2018–2019 Otto Kekäläinen <otto@kekalainen.net>
#    Copyright (C) 2021 Christian Boltz
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import os
import struct
import sqlite3
import re

from apparmor.common import AppArmorBug, DebugLogger

debug_logger = DebugLogger('apparmor.notify')


def get_last_login_timestamp(username, filename='/var/log/wtmp', lastlog2_db='/var/lib/lastlog/lastlog2.db'):
    """Get last login for user as epoch timestamp"""

    if os.access(lastlog2_db, os.R_OK):
        return get_last_login_timestamp_lastlog2(username, lastlog2_db)
    elif os.access(filename, os.R_OK):
        return get_last_login_timestamp_wtmp(username, filename)
    else:
        return 0


def get_last_login_timestamp_lastlog2(username, lastlog2_db='/var/lib/lastlog/lastlog2.db'):
    """Execute lastlog2 and get last login for user as epoch timestamp"""

    db = sqlite3.connect('file:%s?mode=ro' % lastlog2_db, uri=True)
    cur = db.cursor()
    timestamp = cur.execute('SELECT Time FROM Lastlog2 WHERE Name == ?;', [username]).fetchone()

    if timestamp:
        return timestamp[0]
    else:
        return 0


def sane_timestamp(timestamp):
    """Check if the given timestamp is in a date range that makes sense for a wtmp file"""

    if timestamp < 946681200:  # 2000-01-01
        return False
    elif timestamp > 2524604400:  # 2050-01-01
        return False

    return True


def get_last_login_timestamp_wtmp(username, filename='/var/log/wtmp'):
    """Directly read wtmp and get last login for user as epoch timestamp"""
    timestamp = 0
    last_login = 0

    debug_logger.debug('Username: %s', username)

    with open(filename, "rb") as wtmp_file:
        offset = 0
        wtmp_filesize = os.path.getsize(filename)
        debug_logger.debug('WTMP filesize: %s', wtmp_filesize)

        if wtmp_filesize < 356:
            return 0  # (nearly) empty wtmp file, no entries

        # detect architecture based on utmp format differences
        wtmp_file.seek(340)  # first possible timestamp position
        timestamp_x86_64  = struct.unpack("<L", wtmp_file.read(4))[0]  # noqa: E221
        timestamp_aarch64 = struct.unpack("<L", wtmp_file.read(4))[0]
        timestamp_s390x   = struct.unpack(">L", wtmp_file.read(4))[0]  # noqa: E221
        debug_logger.debug('WTMP timestamps: x86_64 %s, aarch64 %s, s390x %s',
                           timestamp_x86_64, timestamp_aarch64, timestamp_s390x)

        if sane_timestamp(timestamp_x86_64):
            endianness = '<'  # little endian
            extra_offset_before = 0
            extra_offset_after = 0
        elif sane_timestamp(timestamp_aarch64):
            endianness = '<'  # little endian
            extra_offset_before = 4
            extra_offset_after = 12
        elif sane_timestamp(timestamp_s390x):
            endianness = '>'  # big endian
            extra_offset_before = 8
            extra_offset_after = 8
        else:
            raise AppArmorBug(
                'Your /var/log/wtmp is broken or has an unknown format. '
                'Please open a bugreport with /var/log/wtmp and the output of "last" attached!')

        while offset < wtmp_filesize:
            wtmp_file.seek(offset)
            offset += 384 + extra_offset_before + extra_offset_after  # Increment for next entry

            type = struct.unpack('%sH' % endianness, wtmp_file.read(2))[0]
            debug_logger.debug('WTMP entry type: %s', type)
            wtmp_file.read(2)  # skip padding

            # Only parse USER lines
            if type == 7:
                # Read each item and move pointer forward
                pid = struct.unpack("<L", wtmp_file.read(4))[0]
                line = wtmp_file.read(32).decode("utf-8", "replace").split('\0', 1)[0]
                id = wtmp_file.read(4).decode("utf-8", "replace").split('\0', 1)[0]
                user = wtmp_file.read(32).decode("utf-8", "replace").split('\0', 1)[0]
                host = wtmp_file.read(256).decode("utf-8", "replace").split('\0', 1)[0]
                term = struct.unpack("<H", wtmp_file.read(2))[0]
                exit = struct.unpack("<H", wtmp_file.read(2))[0]
                session = struct.unpack("<L", wtmp_file.read(4))[0]
                if extra_offset_before:
                    wtmp_file.read(extra_offset_before)
                timestamp = struct.unpack('%sL' % endianness, wtmp_file.read(4))[0]
                if extra_offset_after:
                    wtmp_file.read(extra_offset_after)
                usec = struct.unpack("<L", wtmp_file.read(4))[0]
                entry = (pid, line, id, user, host, term, exit, session, timestamp, usec)
                debug_logger.debug('WTMP entry: %s', entry)

                # Store login timestamp for requested user
                if user == username:
                    last_login = timestamp

    # When loop is done, last value should be the latest login timestamp
    return last_login


def is_special_profile_userns(ev, special_profiles):
    if 'comm' not in ev:
        return False  # special profiles have a 'comm' entry

    if not special_profiles or not special_profiles.match(ev['profile']):
        return False  # We don't use special profiles or there is already a profile defined: we don't ask to add userns

    return True


def get_event_special_type(ev, special_profiles):
    if is_special_profile_userns(ev, special_profiles):
        if ev['operation'] == 'userns_create':
            if ev['aamode'] == 'DENIED':
                return 'userns_denied'
            else:
                return 'userns_change_profile'
        elif ev['operation'] == 'change_onexec':
            return 'userns_change_profile'
        elif ev['operation'] == 'capable':
            return 'userns_capable'
        else:
            raise AppArmorBug('unexpected operation: %s' % ev['operation'])
    return 'normal'


def set_userns_special_profile(special_profiles):
    return re.compile('^({})$'.format('|'.join(special_profiles)))
